#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Parsing utilities for extracting dependencies from SQL Alchemy select statements."""
from logging import getLogger
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

from sqlalchemy.orm.util import _ORMJoin
from sqlalchemy.schema import Table as SATable
from sqlalchemy.sql.elements import (
    BinaryExpression,
    BindParameter,
    ClauseList,
    ColumnClause,
    ColumnElement,
    TextClause,
)
from sqlalchemy.sql.expression import Alias, BooleanClauseList, Function, Label  # type: ignore
from sqlalchemy.sql.expression import Select as _Select
from sqlalchemy.sql.schema import Column as SAColumn

from dbgen.exceptions import QueryParsingError
from dbgen.utils.lists import flatten

# TODO Allow for arrow expressions
logger = getLogger('dbgen.statement_parsing')


def expand_clause_list(clause_statement):
    output = []
    if getattr(clause_statement, "clauses", None) is not None:
        for element in getattr(clause_statement, "clauses"):
            if getattr(clause_statement, "clauses", None) is not None:
                output.extend(expand_clause_list(element))
            else:
                output.append(element)
    else:
        return [clause_statement]
    return output


def get_where_dependency(where_stmt: BooleanClauseList):
    tables = []
    columns = []
    fks = []
    for child in where_stmt.get_children():
        if isinstance(child, SATable):
            tables.append(child)
        elif isinstance(child, SAColumn) and child.foreign_keys:
            fks.append(child)
        elif isinstance(child, SAColumn):
            columns.append(expand_col(child))
        elif isinstance(child, (_ORMJoin, ColumnElement)):
            c_tables, c_columns, c_fks = get_from_dependency(child)
            tables.extend(c_tables)
            columns.extend(c_columns)
            fks.extend(c_fks)
        else:
            print(type(child))
    return columns, tables, fks


def expand_col(column: ColumnElement):
    if isinstance(column, SAColumn):
        if isinstance(column.table, Alias):
            if isinstance(column.table.original, SATable):
                return [column.table.original.columns.get(column.name)]
            else:
                raise NotImplementedError(column.table.original, type(column.table.original))
        return [column]
    elif isinstance(column, Label):
        return expand_col(column.element)
    elif isinstance(column, BinaryExpression):
        return [*expand_col(column.left), *expand_col(column.right)]
    elif isinstance(column, (BindParameter, TextClause)):
        return []
    elif getattr(column, 'get_children', None):
        return list(flatten(map(expand_col, column.get_children())))
    else:
        raise NotImplementedError(f"Unknown selected column type:\n{column}\n{type(column)}")


def get_select_dependency(select_stmt: _Select):
    selected = set(select_stmt.selected_columns)
    columns = set()
    for col in selected:
        columns.update(expand_col(col))

    if select_stmt._having_criteria is not None:  # type: ignore
        for criteria in select_stmt._having_criteria:  # type: ignore
            columns.update(expand_col(criteria))

    tables = {column.table for column in columns}
    fks = {column for column in columns if column.foreign_keys}

    return columns, tables, fks


def get_from_dependency(from_statement: Union[_ORMJoin, ColumnElement, ClauseList]):
    tables = set()
    columns = set()
    fks = set()
    for child in from_statement.get_children():
        if isinstance(child, Alias):
            tables.add(child.original)
        elif isinstance(child, SATable):
            tables.add(child)
        elif isinstance(child, SAColumn) and child.foreign_keys:
            fks.add(child)
        elif isinstance(child, SAColumn):
            columns.update(expand_col(child))
        elif isinstance(child, (_ORMJoin, ColumnElement, ClauseList)):
            c_columns, c_tables, c_fks = get_from_dependency(child)
            tables.update(c_tables)
            columns.update(c_columns)
            fks.update(c_fks)
        else:
            print(type(child))
    return columns, tables, fks


def get_statement_dependency(
    select_stmt: _Select,
) -> Tuple[Set, Set, Set]:
    """Parses a sqlalchemy select statement to get its dependencies.

    Parses the select, where, order_by, and group by clause for all properties, tables, and foreign keys
    required to run the query.

    Args:
        select_stmt (_Select): A SQL alchemy select statement

    Returns:
        Tuple[Set[str],Set[str],Set[str]]: 3 sets of the column, table, and fk dependencies
    """
    # Parse the select clause
    select_cols, select_tabs, select_fks = get_select_dependency(select_stmt)
    # Parse the where clause
    if select_stmt.whereclause is not None:
        where_cols, where_tabs, where_fks = get_from_dependency(select_stmt.whereclause)
    else:
        where_cols, where_tabs, where_fks = set(), set(), set()

    from_cols, from_fks, from_tabs = set(), set(), set()
    # compile the from statements
    from_statements = select_stmt.get_final_froms()  # type: ignore
    if len(from_statements) > 1:
        logger.debug("Warning: 2 from statements detected possible crossjoin used")
    # Add the order_by clause if set
    if select_stmt._order_by_clause is not None:  # type: ignore
        from_statements.append(select_stmt._order_by_clause)  # type: ignore
    # Add the group_by clause if set
    if select_stmt._group_by_clause is not None:  # type: ignore
        from_statements.append(select_stmt._group_by_clause)  # type: ignore

    for from_statement in from_statements:
        c_from_cols, c_from_tabs, c_from_fks = get_from_dependency(from_statement)
        from_cols.update(c_from_cols)
        from_tabs.update(c_from_tabs)
        from_fks.update(c_from_fks)

    # merge dependencies
    cols = select_cols | where_cols | from_cols
    tabs = select_tabs | where_tabs | from_tabs
    fks = select_fks | where_fks | from_fks
    return cols, tabs, fks


def _get_select_keys(select_statement: _Select) -> List[str]:
    output_keys: Dict[str, Optional[str]] = {}

    for column in select_statement.selected_columns:
        col_key, marker = _parse_column(column)
        if col_key in output_keys:
            raise ValueError(
                f"Ambiguous Column Name: Both {marker} and {output_keys[col_key]} have column {col_key}.\n"
                "Please alias one of them using something like col.label('alias')"
            )
        output_keys[col_key] = marker
    return list(output_keys.keys())


def _parse_column(column: Any) -> Tuple[str, Optional[str]]:

    try:
        if isinstance(column, SAColumn):
            col_key = column.name
            marker = f"{column.table.name}.{col_key}"
            return (col_key, marker)
        elif isinstance(column, Alias):
            col_key = column.name  # type: ignore
            marker = f"{column.table.name}.{col_key}"  # type: ignore
            return (col_key, marker)
        elif isinstance(column, Label):
            col_key = column.key
            expanded_col = expand_col(column)
            marker = expanded_col
            return (col_key, marker)
        elif isinstance(column, ColumnClause):
            expanded_col = expand_col(column)
            if len(expanded_col) == 0:
                if isinstance(str, column.name):
                    column.name = cast(str, column.name)  # type: ignore
                    return (column.name, None)
                raise QueryParsingError(f"Unknown column name: {column}")
        elif isinstance(column, (Function, BinaryExpression)):
            raise QueryParsingError(
                "SQLAlchemy Functions need to be labelled due to the imprecise naming of function columns in sqlalchemy.\n"
                f"Try something like \"{getattr(column,'name','COLUMN_NAME')}(YOUR_COLUMN).label('my_column_name')\""
            )
    except QueryParsingError:
        raise
    except AttributeError as exc:
        raise QueryParsingError(
            "Error occurred during parsing that led to an internal sqlalchemy error\n" f"Column: {column}"
        ) from exc

    raise NotImplementedError(str(column), type(column))



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import inspect
from functools import partial
from typing import Callable, Generic, List, Optional, Tuple, TypeVar, Union, overload

from pydantic import ValidationError
from pydantic.typing import get_args, get_origin
from typing_extensions import ParamSpec

from dbgen.core.args import Arg
from dbgen.core.func import Environment
from dbgen.core.node.computational_node import ComputationalNode
from dbgen.core.node.extract import PythonExtract
from dbgen.core.node.transforms import PythonTransform
from dbgen.exceptions import InvalidArgument

In = ParamSpec('In')
Out = TypeVar('Out')
T = TypeVar('T')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')


class FunctionNode(Generic[In, Out]):
    """Temporary wrapper API for pyblocks."""

    def __init__(
        self,
        *inputs,
        function: Callable[In, Out] = None,
        env: Optional[Environment] = None,
        outputs=None,
        node: Optional[ComputationalNode] = None,
    ):
        self.function = function
        self.env = env
        self.inputs = inputs or []
        self.outputs = outputs or ['out']
        self.node = node
        self._arglist = tuple(iter(map(node.__getitem__, self.node.outputs)))

    def __call__(self: 'FunctionNode[In,Out]', *args: In.args, **kwargs: In.kwargs) -> Out:
        return self.function(*args, **kwargs)

    def __getitem__(self: 'FunctionNode[In,Out]', key: Union[str, int]) -> Arg:
        if isinstance(key, str):
            return self.node[key]
        return self._arglist[key]

    @overload
    def results(self: 'FunctionNode[In,Tuple[T1]]') -> Tuple[Arg[T1]]:
        ...  # pragma: no cover

    @overload
    def results(self: 'FunctionNode[In,Tuple[T1,T2]]') -> Tuple[Arg[T1], Arg[T2]]:
        ...  # pragma: no cover

    @overload
    def results(self: 'FunctionNode[In,Tuple[T1,T2,T3]]') -> Tuple[Arg[T1], Arg[T2], Arg[T3]]:
        ...  # pragma: no cover

    @overload
    def results(self: 'FunctionNode[In,T1]') -> Arg[T1]:
        ...  # pragma: no cover

    @overload
    def results(
        self: 'FunctionNode[In,Tuple[T1,T2,T3,T4]]',
    ) -> Tuple[Arg[T1], Arg[T2], Arg[T3], Arg[T4]]:
        ...  # pragma: no cover

    def results(self):
        return self._arglist[0] if len(self._arglist) == 1 else self._arglist


class TransformNode(FunctionNode[In, Out]):
    """Temporary wrapper API for pyblocks."""

    def __init__(
        self,
        *inputs,
        function: Callable[In, Out] = None,
        env: Optional[Environment] = None,
        outputs=None,
    ):
        outputs = outputs or ['out']
        try:
            node = PythonTransform(function=function, env=env, inputs=inputs, outputs=outputs)
        except ValidationError as exc:
            raise InvalidArgument(
                f'Error occurred during the validation of the transform {function.__name__!r}'
            ) from exc
        super().__init__(*inputs, function=function, env=env, outputs=outputs, node=node)


class ExtractNode(FunctionNode[In, Out]):
    """Temporary wrapper API for pyblocks."""

    def __init__(
        self,
        *inputs,
        function: Callable[In, Out] = None,
        env: Optional[Environment] = None,
        outputs=None,
    ):
        outputs = outputs or ['out']

        try:
            node = PythonExtract(function=function, env=env, inputs=inputs, outputs=outputs)
        except ValidationError as exc:
            raise InvalidArgument(
                f'Error occurred during the validation of the transform {function.__name__!r}'
            ) from exc
        super().__init__(*inputs, function=function, env=env, outputs=outputs, node=node)


@overload
def transform(function: Callable[In, Out]) -> Callable[In, FunctionNode[In, Out]]:
    ...  # pragma: no cover


@overload
def transform(
    *, env: Environment = None, outputs: List[str] = None
) -> Callable[[Callable[In, Out]], Callable[In, FunctionNode[In, Out]]]:
    ...  # pragma: no cover


def transform(function=None, *, env: Optional[Environment] = None, outputs: List[str] = None):

    if function:
        if not outputs:

            # TODO add dict outputs or list of dict outputs!
            sig = inspect.signature(function)
            if outputs is None and sig.return_annotation:
                annotation = sig.return_annotation
                origin = get_origin(annotation)
                if origin is not None and origin is not Union and issubclass(origin, (list, tuple)):
                    args = get_args(annotation)
                    bad_args = list(filter(lambda x: not isinstance(x, type), args))
                    if not bad_args:
                        outputs = [str(i) for i, _ in enumerate(args)]
        func = partial(TransformNode, function=function, env=env, outputs=outputs)

        def set_inputs(*inputs: List[Arg]) -> FunctionNode[In, Out]:
            return func(*inputs)

        return set_inputs
    else:
        return partial(transform, env=env, outputs=outputs)


@overload
def extract(function: Callable[In, Out]) -> Callable[In, FunctionNode[In, Out]]:
    ...  # pragma: no cover


@overload
def extract(
    *, env: Environment = None, outputs: List[str] = None
) -> Callable[[Callable[In, Out]], Callable[In, FunctionNode[In, Out]]]:
    ...  # pragma: no cover


def extract(function=None, *, env: Optional[Environment] = None, outputs: List[str] = None):

    if function:
        if not outputs:

            # TODO add dict outputs or list of dict outputs!
            sig = inspect.signature(function)
            if outputs is None and sig.return_annotation:
                annotation = sig.return_annotation
                origin = get_origin(annotation)
                if origin is not None and origin is not Union and issubclass(origin, (list, tuple)):
                    args = get_args(annotation)
                    bad_args = list(filter(lambda x: not isinstance(x, type), args))
                    if not bad_args:
                        outputs = [str(i) for i, _ in enumerate(args)]
        func = partial(ExtractNode, function=function, env=env, outputs=outputs)

        def set_inputs(*inputs: List[Arg]) -> ExtractNode[In, Out]:
            return func(*inputs)

        return set_inputs
    else:
        return partial(extract, env=env, outputs=outputs)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from traceback import format_exc
from typing import TYPE_CHECKING, Any, Callable, Dict, Mapping, Optional, TypeVar, Union

from pydantic import Field, root_validator, validator

from dbgen.configuration import config
from dbgen.core.func import Environment, Func, func_from_callable
from dbgen.core.node.computational_node import ComputationalNode
from dbgen.exceptions import DBgenExternalError, DBgenPythonTransformError, DBgenSkipException
from dbgen.utils.log import capture_stdout

if TYPE_CHECKING:
    from dbgen.core.run.utilities import RunConfig
Output = TypeVar('Output')


class Transform(ComputationalNode[Output]):
    pass


_INJECTED_KWARGS = {'settings'}

# TODO add better error messaging when user passes in a non-arg to pyblock
class PythonTransform(Transform[Output]):

    env: Optional[Environment] = Field(default_factory=lambda: Environment(imports=set()))
    function: Func[Output]

    @validator('function', pre=True)
    def convert_callable_to_func(cls, function: Union[Func[Output], Callable[..., Output]], values):
        env = values.get('env', Environment())
        if isinstance(function, (Func, dict)):
            return function
        elif callable(function):
            return func_from_callable(function, env=env)
        raise ValueError(f"Unknown function type {type(function)} {function}")

    @root_validator
    def check_nargs(cls, values):
        if "function" not in values:
            return values
        inputs = values.get("inputs")
        func: Func = values.get("function")
        if not func:
            raise ValueError('Function is set to None, a previous validation has likely failed')
        num_req_args = func.number_of_required_inputs
        num_max_args = func.number_of_inputs

        number_of_injected_kwargs = len(list(filter(lambda x: x in func.argnames, _INJECTED_KWARGS)))
        num_req_args -= number_of_injected_kwargs
        number_of_inputs = len(inputs) if inputs else 0
        assert (
            number_of_inputs >= num_req_args
        ), f"Too few arguments supplied to Func. Number of Inputs: {number_of_inputs}, Number of Args: {num_req_args}\n{values}"
        if not func.var_positional_keyword:
            assert (
                number_of_inputs <= num_max_args
            ), f"Too many arguments supplied to Func. Number of Inputs: {number_of_inputs}, Max Number of Args: {num_max_args}\n"
        return values

    def run(
        self, namespace_dict: Dict[str, Mapping[str, Any]], run_config: Optional['RunConfig'] = None
    ) -> Dict[str, Any]:
        inputvars = self._get_inputs(namespace_dict)
        args = {key: val for key, val in inputvars.items() if key.isdigit()}
        kwargs = {key: val for key, val in inputvars.items() if key not in args}

        if 'settings' in self.function.argnames and run_config:
            setting_index = self.function.argnames.index('settings')
            if 'settings' not in kwargs or len(args) < setting_index + 1:
                kwargs['settings'] = run_config.settings

        try:
            wrapped = capture_stdout(self.function) if not config.pdb else self.function
            output = wrapped(*args.values(), **kwargs)
            if isinstance(output, tuple):
                l1, l2 = len(output), len(self.outputs)
                assert l1 == l2, "Expected %d outputs from %s, got %d" % (
                    l2,
                    self.function.name,
                    l1,
                )
                return {o: val for val, o in zip(output, self.outputs)}
            else:
                if len(self.outputs) != 1:
                    raise DBgenPythonTransformError(
                        f"Function returned a non-tuple but outputs is greater length 1: {self.outputs}"
                    )
                return {list(self.outputs)[0]: output}
        except (DBgenSkipException, DBgenPythonTransformError):
            raise
        except Exception:
            msg = f"Error encountered while applying function named {self.function.name!r}"
            self._logger.exception(format_exc())
            raise DBgenExternalError(msg)

    def __call__(self, *args, **kwargs) -> Output:
        return self.function(*args, **kwargs)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections.abc import Iterable
from functools import reduce
from itertools import repeat
from typing import Dict, Generator, Sequence, Set, Tuple, TypeVar, overload

##############################################################
T = TypeVar("T")
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")


def is_broadcastable(*args) -> bool:
    """
    Enforce that all non-length-1 elements have the same length
    """
    length_set: Set[int] = reduce(lambda x, y: x.union([len(y)]), args, set())
    # Check lengths
    error_message = f"Found {len(length_set)-(1 in length_set)} different >1-length iterables, cannot broadcast: \nLengths: {length_set}"
    if not len(length_set) <= 2 or (len(length_set) > 1 and 1 not in length_set and 0 not in length_set):
        raise ValueError(error_message)
    return True


@overload
def broadcast(
    __vals1: Sequence[_T1],
    __vals2: Sequence[_T2],
) -> Generator[Tuple[_T1, _T2], None, None]:
    ...


@overload
def broadcast(
    __vals1: Sequence[_T1],
    __vals2: Sequence[_T2],
    __vals3: Sequence[_T3],
) -> Generator[Tuple[_T1, _T2, _T3], None, None]:
    ...


@overload
def broadcast(
    __vals1: Sequence[_T1],
    __vals2: Sequence[_T2],
    __vals3: Sequence[_T3],
    __vals4: Sequence[_T4],
) -> Generator[Tuple[_T1, _T2, _T3, _T4], None, None]:
    ...


@overload
def broadcast(
    __vals1: Sequence[_T1],
    __vals2: Sequence[_T2],
    __vals3: Sequence[_T3],
    __vals4: Sequence[_T4],
    __vals5: Sequence[_T5],
) -> Generator[Tuple[_T1, _T2, _T3, _T4, _T5], None, None]:
    ...


def broadcast(*args: Sequence[T]) -> Generator[Tuple[T, ...], None, None]:
    """Broadcast list of sequences together, requires all lists to be same length or length 1.

    Yields:
        Generator[Tuple[T, ...], None, None]: Generator that returns a tuple from each input Sequence
    """
    # Check that args are broadcastable
    is_broadcastable(*args)
    iterators = [iter(arg) for arg in args]
    num_active = len(iterators)
    empty_lists = [l for l in args if len(l) == 0]
    if empty_lists:
        # If there are empty lists return N empty lists
        yield from ([] for _ in range(num_active))  # type: ignore
        return
    fillvalue: Dict[int, T] = {}
    if not num_active:
        return
    while True:
        values = []
        for i, it in enumerate(iterators):
            try:
                value = next(it)
                fillvalue[i] = value
            except StopIteration:
                try:
                    num_active -= 1
                    if not num_active:
                        return
                    iterators[i] = repeat(fillvalue[i])
                    value = fillvalue[i]
                except KeyError:
                    raise ValueError(f"Error broadcasting lists: {args}")
            values.append(value)
        yield tuple(values)


def flatten(thing):
    for element in thing:
        if isinstance(element, Iterable) and not isinstance(element, str):
            yield from flatten(element)
        else:
            yield element



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING
from typing import Generator as GenType
from typing import List, Optional, Tuple
from uuid import UUID

import typer
from sqlmodel import Session, func, select

from dbgen.configuration import get_engines, stdout_handler
from dbgen.core.metadata import ETLStepEntity, ETLStepRunEntity, RunEntity
from dbgen.core.run.etl_step_run import RemoteETLStepRun
from dbgen.core.run.utilities import RunConfig
from dbgen.utils.log import LogLevel

if TYPE_CHECKING:
    from sqlalchemy.future import Engine  # pragma: no cover

etl_step_app = typer.Typer()


def get_runnable_etl_steps(
    all_runs: bool = True, meta_engine: Optional['Engine'] = None
) -> GenType[Tuple[str, datetime, UUID], None, None]:
    statement = (
        select(  # type: ignore
            ETLStepEntity.name,
            func.max(ETLStepRunEntity.created_at),
            ETLStepEntity.id,
        )
        .join_from(ETLStepEntity, ETLStepRunEntity)
        .group_by(ETLStepEntity.name, ETLStepEntity.id)
        .order_by(ETLStepEntity.name, func.max(ETLStepRunEntity.created_at).desc())
    )
    if not all_runs:
        statement = statement.where(
            ETLStepRunEntity.run_id == select(func.max(RunEntity.id)).scalar_subquery()  # type: ignore
        )

    # If we don't have a metaengine grab one
    if not meta_engine:
        _, meta_engine = get_engines()

    with Session(meta_engine) as session:
        result = session.exec(statement.distinct())
        yield from result


@etl_step_app.command('list')
def list_etl_steps(
    all_runs: bool = typer.Option(
        False, '-a', '--all', help='Print all possible ETLSteps not just most recent'
    )
):
    """List out the ETLSteps from the last run."""
    typer.echo("-" * 77)
    typer.echo(f"{'name':<20} {'last_run':<20} {'etl_step_id':20}")
    typer.echo("-" * 77)
    for etl_step_name, last_run, etl_step_id in get_runnable_etl_steps(all_runs):
        typer.echo(
            f"{etl_step_name!r:<20} {last_run.strftime('%m/%d/%Y %H:%M:%S'):<20} {str(etl_step_id):<20}"
        )


@etl_step_app.command('run')
def run_etl_step(
    etl_step_id: UUID,
    run_id: Optional[int] = typer.Argument(None),
    ordering: Optional[int] = typer.Argument(None),
    retry: bool = typer.Option(False, help="Ignore repeat checking"),
    level: LogLevel = typer.Option(LogLevel.INFO, help="Log level"),
    include: List[str] = typer.Option([], help="ETLSteps to include"),
    exclude: List[str] = typer.Option([], help="ETLSteps to xclude"),
    start: Optional[str] = typer.Option(None, help="ETLStep to start run at"),
    until: Optional[str] = typer.Option(None, help="ETLStep to finish run at."),
    batch: Optional[int] = typer.Option(None, help="Batch size for all etl_steps"),
):
    # Retrieve the configured engines for
    main_engine, meta_engine = get_engines()
    run_config: RunConfig = RunConfig(
        retry=retry,
        start=start,
        until=until,
        exclude=exclude,
        include=include,
        progress_bar=False,
        batch_size=batch,
        log_level=level,
    )
    stdout_handler.setLevel(level.get_log_level())
    # Validate that the etl_step id provided is in the ETLStepEntity Table for remote running
    etl_steps = {
        etl_step_id: etl_step_name
        for etl_step_name, _, etl_step_id in get_runnable_etl_steps(meta_engine=meta_engine)
    }
    if etl_step_id not in etl_steps:
        raise typer.BadParameter(f"Invalid etl_step id, no etl_step found with uuid {etl_step_id}.")

    # Need to validate the run_id to make sure that run has been initialized (i.e. a run exists with run_id)
    # and that a etl_step_run has not already been created with this etl_step_id and run_id
    if run_id is not None:
        with Session(meta_engine) as session:
            existing_etl_step_run = session.get(ETLStepRunEntity, (etl_step_id, run_id))
            if existing_etl_step_run:
                raise typer.BadParameter(
                    f"Found an existing etl_step_run with (etl_step_id, run_id) equal to ({etl_step_id},{run_id})"
                )
            run = session.get(RunEntity, run_id)
            if not run:
                raise typer.BadParameter(
                    f"No Run exists with run_id = {run_id} has the run been initialized?"
                )
    code = RemoteETLStepRun(etl_step_id=etl_step_id).execute(
        main_engine, meta_engine, run_id, run_config, ordering
    )
    if code:
        raise typer.Exit(code=code)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
from typing import Optional

reserved_words = {
    "and",
    "del",
    "for",
    "is",
    "raise",
    "assert",
    "elif",
    "from",
    "lambda",
    "return",
    "break",
    "else",
    "global",
    "not",
    "try",
    "class",
    "except",
    "if",
    "or",
    "while",
    "continue",
    "exec",
    "import",
    "pass",
    "yield",
    "def",
    "finally",
    "in",
    "print",
    "as",
}


def which(program: str) -> Optional[str]:
    def is_exe(fpath: str) -> bool:
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ.get("PATH", '').split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Methods related to fast copying to a postgresql database."""
import asyncio
import re
from logging import getLogger
from typing import TYPE_CHECKING, Any, Iterable, Mapping, Optional, Sequence, Tuple
from uuid import UUID

from psycopg.errors import UndefinedColumn

from dbgen.exceptions import DatabaseError

if TYPE_CHECKING:
    from psycopg import AsyncConnection, Connection

    from dbgen.core.node.load import LoadEntity

# Error messages
MISSING_COLUMN_ERROR = "An ETLStep is trying to load into the entity {name!r}, however {column} does not exist. This commonly occurs when a column is added to an entity without being effectively synced with the database. An easy fix is to rerun the model with --build, if this is not possible you can manually add the column. If this column "
MISSING_ID_ERROR = "An ETLStep is trying to load into the entity {name!r}, however {column} does not exist. This can occur when the entities in the model comes out of sync with the database. Since {column} is an identifying column of {name!r} the id hashes for the rows in the table will change when adding this column therefore you must rebuild the database with --build."

# SQL Statements for the get_statements function
CREATE_TABLE_STATEMENT = """
CREATE TEMPORARY TABLE {temp_table_name} AS
TABLE {full_table_name}
WITH NO DATA;
ALTER TABLE {temp_table_name}
ADD COLUMN auto_inc SERIAL NOT NULL;
"""

UPDATE_STATEMENT = """
UPDATE
    {full_table_name}
SET
    {column_str}
FROM
    {temp_table_name}
WHERE
    {full_table_name}.{table_primary_key_name} = {temp_table_name}.{table_primary_key_name};
"""

INSERT_STATEMENT = """
INSERT INTO {full_table_name}
({all_columns_str})
SELECT
{all_columns_str}
FROM (
  SELECT
    {all_columns_str},
    ROW_NUMBER() OVER (PARTITION BY {table_primary_key_name}
               ORDER BY auto_inc {order}) AS row_number
  FROM
    {temp_table_name}) AS X
WHERE
  row_number = 1 ON CONFLICT {conflict_key}
  DO
  UPDATE
  SET
  {update_column_statement}
  RETURNING
    {table_primary_key_name}
"""


def escape_str(column: str):
    return f"\"{column}\""


def _get_types(load_entity: 'LoadEntity', columns):
    oids = list(map(lambda x: load_entity.attributes[x], sorted(columns)))
    return oids


def load_data(
    data: Mapping[UUID, Sequence[Any]],
    connection: 'Connection',
    load_entity: 'LoadEntity',
    columns: Iterable[str],
    insert: bool,
    temp_table_suffix: str = '',
    etl_step_id: Optional[UUID] = None,
) -> int:
    # Setup the logger
    logger = getLogger(f'dbgen.load.{load_entity.name}')
    # Get the SQL Statements for this load
    create_statement, drop_statement, copy_statement, load_statement = get_statements(
        load_entity.name,
        load_entity.full_name,
        load_entity.primary_key_name,
        insert,
        columns,
        temp_table_suffix=temp_table_suffix,
        etl_step_id=etl_step_id,
    )
    # Drop the Create the Temporary table
    logger.debug('creating temp table')
    with connection.cursor() as cur:
        cur.execute(drop_statement)
        cur.execute(create_statement)
    connection.commit()

    with connection.cursor() as cur:
        logger.debug("load into temporary table")
        try:
            with cur.copy(copy_statement) as copy:
                oids = _get_types(load_entity, [load_entity.primary_key_name] + list(columns))
                copy.set_types(oids)
                for pk_curr, row_curr in data.items():
                    copy.write_row((pk_curr, *row_curr))
        except UndefinedColumn as exc:
            # Try to match the column name to give helpful error messages
            match = re.match('column \"(\\w+)\"', str(exc))
            if match:
                (column,) = match.groups()
                column_str = f'the column {column!r}'
            else:
                column_str = 'a column'

            if column in load_entity.identifiers:
                raise DatabaseError(MISSING_ID_ERROR.format(name=load_entity.name, column=column_str))
            raise DatabaseError(
                MISSING_COLUMN_ERROR.format(name=load_entity.name, column=column_str)
            ) from exc
        # Try to insert everything from the temp table into the real table
        # If a foreign_key violation is hit, we delete those rows in the
        # temp table and move on
        logger.debug("transfer from temp table to main table")
        cur.execute(load_statement)
        logger.debug("Dropping temp table...")
        cur.execute(drop_statement)
    connection.commit()
    logger.debug("loading finished")
    return len(data)


async def async_load_data(
    data: Mapping[UUID, Sequence[Any]],
    connection: 'AsyncConnection',
    load_entity: 'LoadEntity',
    columns: Iterable[str],
    insert: bool,
    temp_table_suffix: str = '',
    etl_step_id: Optional[UUID] = None,
) -> int:
    # Setup the logger
    logger = getLogger(f'dbgen.async_load.{load_entity.name}')
    task = asyncio.current_task()
    # Get the SQL Statements for this load
    create_statement, drop_statement, copy_statement, load_statement = get_statements(
        load_entity.name,
        load_entity.full_name,
        load_entity.primary_key_name,
        insert,
        columns,
        temp_table_suffix=f'{temp_table_suffix}_{hash(task)}',
        etl_step_id=etl_step_id,
    )
    # Drop the Create the Temporary table
    logger.debug('creating temp table')
    async with connection.cursor() as cur:
        await cur.execute(drop_statement)
        logger.debug(create_statement)
        await cur.execute(create_statement)
    await connection.commit()

    async with connection.cursor() as cur:
        logger.debug("load into temporary table")
        try:
            async with cur.copy(copy_statement) as copy:
                oids = _get_types(load_entity, [load_entity.primary_key_name] + list(columns))
                copy.set_types(oids)
                for pk_curr, row_curr in data.items():
                    await copy.write_row((pk_curr, *row_curr))
        except UndefinedColumn as exc:
            # Try to match the column name to give helpful error messages
            match = re.match('column \"(\\w+)\"', str(exc))
            if match:
                (column,) = match.groups()
                column_str = f'the column {column!r}'
            else:
                column_str = 'a column'

            if column in load_entity.identifiers:
                raise DatabaseError(MISSING_ID_ERROR.format(name=load_entity.name, column=column_str))
            raise DatabaseError(
                MISSING_COLUMN_ERROR.format(name=load_entity.name, column=column_str)
            ) from exc
        # Try to insert everything from the temp table into the real table
        # If a foreign_key violation is hit, we delete those rows in the
        # temp table and move on
        logger.debug("transfer from temp table to main table")
        await cur.execute(load_statement)
        logger.debug("Dropping temp table...")
        await cur.execute(drop_statement)
    await connection.commit()
    logger.debug("loading finished")
    return len(data)


def get_statements(
    table_name: str,
    full_table_name: str,
    table_primary_key_name: str,
    insert: bool,
    columns: Iterable[str],
    etl_step_id: Optional[UUID] = None,
    temp_table_suffix: str = '',
    partition_attribute: Optional[str] = None,
) -> Tuple[str, str, str, str]:
    """
    Generate the SQL statements relevant for bulk loading data into postgresql.sql

    Args:
        table_name (str): The table name to load into postgresql.sql
        schema (str): The schema of the table to load into postgresql.sql
        load_entity_hash (str): A unique hash for this loads
        insert (bool): Whether or not the statement should be update or inserted
        etl_step_id (UUID): The ETLStep UUID that is loading this data.sql

    Returns:
        Tuple[str, str, str]: The create_table, drop_table, and load statements
    """
    # create
    sanitized_name = full_table_name.replace(".", "_").replace('"', '')
    temp_table_name = f'{sanitized_name}_temp_load_table'
    if temp_table_suffix:
        temp_table_name += f'_{temp_table_suffix}'
    temp_table_name = escape_str(temp_table_name)
    all_columns = [escape_str(table_primary_key_name)] + list(sorted(map(escape_str, columns)))

    create_statement = CREATE_TABLE_STATEMENT.format(
        temp_table_name=temp_table_name,
        table_name=table_name,
        full_table_name=full_table_name,
        etl_step_id=etl_step_id,
    )
    drop_statement = f"DROP TABLE IF EXISTS {temp_table_name}"
    # The Copy statement does not need to insert the ETLStep ID
    copy_columns_str = ', '.join(all_columns)
    copy_statement = f'COPY  {temp_table_name} ({copy_columns_str}) FROM STDIN'
    # If an etl_step_id is provided add it as a default column on temp table so each row
    # modified by these statements track their id
    if etl_step_id is not None:
        create_statement += (
            f"ALTER TABLE {temp_table_name} ALTER COLUMN \"etl_step_id\" SET DEFAULT '{etl_step_id}'"
        )
        all_columns.append(escape_str('etl_step_id'))

    column_statement = [f'{column} = {temp_table_name}.{column}' for column in all_columns]
    first = False
    full_kwargs = dict(
        temp_table_name=temp_table_name,
        table_primary_key_name=escape_str(table_primary_key_name),
        full_table_name=full_table_name,
        all_columns_str=', '.join(all_columns),
        column_str=', '.join(column_statement),
        order='ASC' if first else 'DESC',
    )

    if insert:
        if partition_attribute:
            full_kwargs['conflict_key'] = f'("{table_primary_key_name}","{partition_attribute}")'
        else:
            full_kwargs['conflict_key'] = f'({escape_str(table_primary_key_name)})'
        update_column_statement = ', '.join([f'{column} = excluded.{column}' for column in all_columns])
        full_kwargs['update_column_statement'] = update_column_statement
        load_statement = INSERT_STATEMENT.format(**full_kwargs)
    else:
        load_statement = UPDATE_STATEMENT.format(**full_kwargs)

    return create_statement, drop_statement, copy_statement, load_statement



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typer import Argument, Option

from dbgen.cli.utils import chdir_callback, config_callback, set_verbosity, version_callback

model_string_option = Option(None, '--model', envvar=["DBGEN_MODEL_STR", "dbgen_model_str"])
model_arg_option = Argument('--model', envvar=["DBGEN_MODEL_STR", "dbgen_model_str"])
version_option = Option(None, "--version", callback=version_callback, is_eager=True)
config_option = Option(
    '.env', "--config", "-c", callback=config_callback, help="Configuration file.", envvar='DBGEN_CONFIG'
)
log_file_option = Option(None, "--log-file", help="File to write logs to")
verbose_option = lambda v=True: Option(
    v, "--verbose", "-v", callback=set_verbosity, help="Increases the verbosity of printed messages"
)
chdir_option = Option(
    None,
    "--chdir",
    callback=chdir_callback,
    is_eager=True,
    help="Change the working directory to find the model",
)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
from typing import Any
from typing import Generator as GenType
from typing import List, Optional, Tuple

import boto3

from dbgen import Extract


class S3FileNameExtractor(Extract[str]):
    bucket: str
    profile: Optional[str]
    prefix: str = ''
    pattern: str = '.*'
    outputs: List[str] = ['s3_key']

    def extract(self) -> GenType[str, None, None]:
        """Return the files in a directory on s3 bucket that match pattern."""
        session = boto3.Session(profile_name=self.profile)
        s3 = session.client("s3")
        regex = re.compile(self.pattern, re.MULTILINE)

        response = s3.list_objects_v2(Bucket=self.bucket, Prefix=self.prefix)
        assert "Contents" in response, response
        if not response or response["KeyCount"] == 0:
            return

        def process_keys(response: dict) -> Tuple[List[str], Any]:
            return (
                [content["Key"] for content in response["Contents"] if regex.search(content["Key"])],
                response.get("NextContinuationToken"),
            )

        vals, token = process_keys(response)
        yield from vals
        count = 0
        while token:
            response = s3.list_objects_v2(Bucket=self.bucket, Prefix=self.prefix, ContinuationToken=token)
            if not response["IsTruncated"]:
                break

            keys, token = process_keys(response)
            yield from keys
            count += response["KeyCount"]


class S3FileExtractor(S3FileNameExtractor):
    """Returns both the key and contents of the file on s3"""

    outputs: List[str] = ['s3_key', 'contents']

    def extract(self) -> GenType[Tuple[str, str], None, None]:  # type: ignore
        session = boto3.Session(profile_name=self.profile)
        s3 = session.resource("s3")
        key_gen = super().extract()
        bucket = s3.Bucket(self.bucket)
        for key in key_gen:
            obj = bucket.Object(key=key)
            body = obj.get()["Body"].read().decode("utf-8")
            yield key, body



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Objects related to the running of Models and ETLSteps."""
from datetime import datetime, timedelta
from time import time
from typing import TYPE_CHECKING

from sqlalchemy.future import Engine
from sqlmodel import Session, select

from dbgen.core.base import Base
from dbgen.core.dashboard import BarNames, Dashboard
from dbgen.core.etl_step import ETLStep
from dbgen.core.metadata import ETLStepsToRun, ModelEntity, RunEntity, Status
from dbgen.core.model import Model
from dbgen.core.run.etl_step_run import (
    AsyncETLStepRun,
    AsyncRemoteETLStepRun,
    BaseETLStepRun,
    ETLStepRun,
    RemoteETLStepRun,
)
from dbgen.core.run.utilities import RunConfig, RunInitializer, update_run_by_id
from dbgen.utils.log import logging_console

if TYPE_CHECKING:
    pass


class ModelRun(Base):
    model: Model

    class Config:
        """Pydantic config"""

        copy_on_model_validation = False

    def get_etl_step_run(self, etl_step: ETLStep, run_async: bool, remote: bool) -> BaseETLStepRun:
        if run_async:
            if remote:
                return AsyncRemoteETLStepRun(etl_step_id=etl_step.uuid)
            else:
                return AsyncETLStepRun(etl_step=etl_step)
        else:
            if remote:
                return RemoteETLStepRun(etl_step_id=etl_step.uuid)
            return ETLStepRun(etl_step=etl_step)

    def execute(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        run_config: RunConfig = None,
        build: bool = False,
        run_async: bool = False,
        remote: bool = False,
        rerun_failed: bool = False,
    ) -> RunEntity:
        start = time()
        if run_config is None:
            run_config = RunConfig()

        run_config.settings = self.model.settings
        # Sync the Database statew with the model state
        self.model.sync(main_engine, meta_engine, build=build)

        # If doing last failed run query for ETLSteps to run and add to include
        if rerun_failed:
            with meta_engine.connect() as conn:
                result = conn.execute(select(ETLStepsToRun.__table__.c.name))
                for (etl_step_name,) in result:
                    run_config.include.add(etl_step_name)

        # Initialize the run
        run_init = RunInitializer()
        run_id = run_init.execute(meta_engine, run_config)
        sorted_etl_steps = self.model._sort_graph()
        # Add etl_steps to metadb
        with Session(meta_engine) as meta_session:
            model_row = self.model._get_model_row()
            model_row.last_run = datetime.now()
            existing_model = meta_session.get(ModelEntity, model_row.id)
            if not existing_model:
                meta_session.merge(model_row)
            else:
                existing_model.last_run = datetime.now()
            meta_session.commit()

        # Apply start and until to exclude etl_steps not between start_idx and until_idx
        if run_config.start or run_config.until:
            etl_step_names = [etl_step.name for etl_step in sorted_etl_steps]
            start_idx = etl_step_names.index(run_config.start) if run_config.start else 0
            until_idx = (
                etl_step_names.index(run_config.until) + 1 if run_config.until else len(etl_step_names)
            )
            # Modify include to only include the etl_step_names that pass the test
            run_config.include = run_config.include.union(etl_step_names[start_idx:until_idx])
            self._logger.debug(
                f"Only running etl_steps: {etl_step_names[start_idx:until_idx]} due to start/until"
            )
        with Dashboard(console=logging_console, enable=run_config.progress_bar).show(
            total=len(sorted_etl_steps)
        ) as dashboard:
            for i, etl_step in enumerate(sorted_etl_steps):
                dashboard.set_etl_name(etl_step.name, i)
                etl_step_run = self.get_etl_step_run(etl_step, run_async, remote)
                code = etl_step_run.execute(
                    main_engine, meta_engine, run_id, run_config, ordering=i, dashboard=dashboard
                )
                # If we fail run exclude downstream generators from running
                if code == 1 and (run_config.fail_downstream or run_config.fast_fail):
                    if run_config.fast_fail:
                        self._logger.info(
                            f'Excluding all downstream ETLSteps due to the failure of {etl_step.name!r}'
                        )
                    for target in sorted_etl_steps[i + 1 :]:
                        if run_config.fast_fail:
                            run_config.upstream_fail_exclude.add(target.name)
                        elif target._get_dependency().test(etl_step._get_dependency()):
                            self._logger.info(
                                f'Excluding ETLStep {target.name!r} due to failed upstream dependency {etl_step.name!r}'
                            )
                            run_config.upstream_fail_exclude.add(target.name)
                elif code == 2:
                    break
                dashboard.advance_bar(BarNames.OVERALL)
            dashboard.finish()

        # Complete run
        run_status = Status.completed if code != 2 else Status.failed
        with Session(meta_engine) as session:
            update_run_by_id(run_id, run_status, session)
            run = session.get(RunEntity, run_id)
            assert run
            run.runtime = timedelta(seconds=time() - start)
            session.commit()
            session.refresh(run)
        return run



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Objects related to the running of Models and ETLSteps."""
from abc import abstractmethod
from typing import TYPE_CHECKING, Dict, List, Optional, Set
from uuid import UUID

from pydantic.fields import Field, PrivateAttr
from sqlalchemy.future import Engine
from sqlmodel import Session

from dbgen.core.base import Base
from dbgen.core.dashboard import Dashboard
from dbgen.core.etl_step import ETLStep
from dbgen.core.metadata import ETLStepRunEntity, RunEntity, Status
from dbgen.core.model import Model
from dbgen.core.model_settings import BaseModelSettings
from dbgen.utils.log import LogLevel

if TYPE_CHECKING:
    pass


class RunConfig(Base):
    """Configuration for the running of an ETLStep and Model"""

    retry: bool = False
    include: Set[str] = Field(default_factory=set)
    exclude: Set[str] = Field(default_factory=set)
    upstream_fail_exclude: Set[str] = Field(default_factory=set)
    start: Optional[str]
    until: Optional[str]
    batch_size: Optional[int]
    progress_bar: bool = True
    skip_row_count: bool = False
    fail_downstream: bool = False
    fast_fail: bool = False
    skip_on_error: bool = False
    batch_number: int = 10
    log_level: LogLevel = LogLevel.INFO
    settings: BaseModelSettings = Field(default_factory=lambda: BaseModelSettings())

    def should_etl_step_run(self, etl_step: ETLStep) -> bool:
        """Check an ETLStep against include/exclude to see if it should run."""
        markers = [etl_step.name, *etl_step.tags]
        should_run = any(
            map(
                lambda x: (not self.include or x in self.include) and x not in self.exclude,
                markers,
            )
        )
        return should_run

    def get_invalid_markers(self, model: Model) -> Dict[str, List[str]]:
        """Check that all inputs to RunConfig are meaningful for the model."""
        invalid_marks = {}
        etl_step_names = model.etl_steps_dict().keys()
        # Validate start and until
        for attr in ("start", "until"):
            val: str = getattr(self, attr)
            if val is not None and val not in etl_step_names:
                invalid_marks[attr] = [val]
        # Validate include and exclude as sets
        for attr in ("include", "exclude"):
            set_val: Set[str] = getattr(self, attr)
            invalid_vals = [x for x in set_val if not model.validate_marker(x)]
            if invalid_vals:
                invalid_marks[attr] = invalid_vals

        return invalid_marks


def update_run_by_id(run_id, status: Status, session: Session):
    run = session.get(RunEntity, run_id)
    assert run, f"No run found with id {run_id}"
    run.status = status
    session.commit()


class RunInitializer(Base):
    """Initializes a run by syncing the database and getting the run_id."""

    def execute(self, engine: Engine, run_config: RunConfig) -> int:
        # Use some metadatabase connection to initialize initialize the run
        # Store the details of the run on the metadatabase so downstream ETLStepRuns can pick them up
        # Sync the database with the registries
        with Session(engine) as session:
            run = RunEntity(status=Status.initialized)
            session.add(run)
            session.commit()
            session.refresh(run)
            assert isinstance(run.id, int)
            run.status = Status.running
            session.commit()
            run_id = run.id

        return run_id


class BaseETLStepExecutor(Base):
    etl_step: ETLStep
    run_config: RunConfig
    _etl_step_run: ETLStepRunEntity = PrivateAttr()
    _old_repeats: Set[UUID] = PrivateAttr(default_factory=set)
    _new_repeats: Set[UUID] = PrivateAttr(default_factory=set)

    @abstractmethod
    def execute(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        meta_session: Session,
        run_id: Optional[int],
        dashboard: Optional[Dashboard],
        etl_step_run: ETLStepRunEntity,
    ) -> int:
        pass



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from contextlib import contextmanager
from enum import Enum
from functools import partial
from typing import Callable, Dict, Tuple

from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.table import Table


class BarNames(str, Enum):
    EXTRACTED = 'Extracted'
    LAUNCHED = 'Launched'
    TRANSFORMED = 'Transformed'
    LOADED = 'Loaded'
    OVERALL = ''


class Dashboard:
    bars: Dict[BarNames, Callable]

    def __init__(self, console: Console = None, enable: bool = True) -> None:
        self.console = console
        self.enable = enable
        self.bars = {}
        self.refresh_per_second = 8

    def make_table(self, progress):
        if not self.enable:
            return
        progress_table = Table.grid(expand=True)
        progress_table.add_row(
            Panel(self.overall_progress, title="Overall Progress", border_style="green", padding=(2, 2)),
            Panel(progress, title="[b]ETL Step", border_style="red", padding=(1, 2)),
        )

        return progress_table

    @property
    def etl_step_progress(self):
        return Progress(
            SpinnerColumn(),
            "[progress.description]{task.description}",
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            "[progress.completed]{task.completed:>3.0f} rows",
            refresh_per_second=self.refresh_per_second,
            transient=True,
        )

    @contextmanager
    def show(self, total: int):
        self.overall_length = total
        self.overall_progress = Progress(TimeElapsedColumn(), BarColumn(), TextColumn("{task.description}"))
        generator_bar = self.overall_progress.add_task(BarNames.OVERALL, total=total)
        self.bars[BarNames.OVERALL] = partial(self.overall_progress.update, generator_bar)
        if not self.enable:
            yield self
            return
        with Live(
            self.make_table(self.etl_step_progress),
            console=self.console,
            refresh_per_second=self.refresh_per_second,
            transient=True,
        ) as live:
            self.live = live
            yield self

    def set_etl_name(self, etl_step_name: str, order: int = None):
        self.advance_bar(
            BarNames.OVERALL,
            advance=0,
            description=f"[bold #AAAAAA]{etl_step_name} ({order}/{self.overall_length})",
        )

    def finish(self):
        self.advance_bar(BarNames.OVERALL, advance=0, description='Finished!')

    def set_total(self, total):
        for bar in self.bars:
            if bar != BarNames.OVERALL:
                self.advance_bar(bar, advance=0, total=total, start=True)

    def add_etl_progress_bars(self, total: int = None, run_async: bool = False):
        if not self.enable:
            return
        progress = self.etl_step_progress
        self.live.update(self.make_table(progress))
        if run_async:
            bar_names: Tuple[BarNames, ...] = (
                BarNames.EXTRACTED,
                BarNames.LAUNCHED,
                BarNames.TRANSFORMED,
                BarNames.LOADED,
            )
        else:
            bar_names = (BarNames.EXTRACTED, BarNames.TRANSFORMED, BarNames.LOADED)
        for text in bar_names:
            bar = progress.add_task(
                text,
                start=total is not None,
                total=total or 1000000,
            )
            self.bars[text] = partial(progress.update, bar)

    def advance_bar(self, bar_name: BarNames, advance: int = 1, description: str = None, **kwargs):
        if self.enable:
            self.bars[bar_name](advance=advance, description=description, **kwargs)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from functools import partial
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from pydantic import Field, PrivateAttr
from pydantic import ValidationError as PydValidationError
from pydantic import root_validator, validate_model, validator
from pydantic.error_wrappers import ErrorWrapper
from pydasher import hasher
from pydasher.import_module import import_string

from dbgen.configuration import ValidationEnum, config
from dbgen.core.args import Arg, Constant
from dbgen.core.base import Base
from dbgen.core.dependency import Dependency
from dbgen.core.node.computational_node import ComputationalNode
from dbgen.core.type_registry import column_registry
from dbgen.exceptions import ValidationError
from dbgen.utils.lists import broadcast, is_broadcastable
from dbgen.utils.postgresql_load import async_load_data, load_data

if TYPE_CHECKING:
    from psycopg import AsyncConnection, Connection  # pragma: no cover

    from dbgen.core.entity import BaseEntity  # pragma: no cover


def hash_tuple(tuple_to_hash: Tuple[Any, ...]) -> UUID:
    return UUID(hasher(tuple_to_hash))


class LoadEntity(Base):
    """Object for passing minimum info to the Load object for database insertion."""

    name: str
    schema_: Optional[str]
    entity_class_str: Optional[str]
    primary_key_name: str
    identifying_attributes: Set[str] = Field(default_factory=dict)
    identifying_foreign_keys: Set[str] = Field(default_factory=set)
    attributes: Dict[str, str] = Field(default_factory=dict)
    required: Set[str] = Field(default_factory=set)
    foreign_keys: Set[str] = Field(default_factory=set)
    _entity: Optional[Type['BaseEntity']] = PrivateAttr(None)

    def __str__(self):
        return (
            "LoadEntity<"
            f"name={self.name!r}, "
            f"schema={self.schema!r}, "
            f"id_attrs={self.identifying_attributes}, "
            f"id_fks={self.identifying_foreign_keys}, "
            f"attributes={self.attributes}, "
            f"required={self.required}, "
            f"fks={self.foreign_keys}>"
        )

    def _load_entity(self) -> Optional[Type['BaseEntity']]:
        if self._entity is None:
            if self.entity_class_str is None:
                self._logger.warning(f"No Entity Class String found cannot validate data!")
                return None
            try:
                entity = import_string(self.entity_class_str)
                self._entity: 'BaseEntity' = entity
            except ImportError:
                self._logger.warning(
                    f"Cannot load entity from class string {self.entity_class_str} cannot validate data!"
                )
                return None
        return self._entity

    # TODO Merge validate and strict validate into one method
    # TODO Decide whether use default is a valid flag for dbgen
    def _validate(self, input_data: Dict[str, Any], use_defaults: bool = False, insert: bool = False):
        entity = self._load_entity()
        if entity is None:
            return input_data
        validated_data, fields_set, errors = validate_model(entity, input_data)
        # If we have errors during validation check if we are inserting
        # if we are not inserting we ignore missing values as updates don't require all info
        # TODO decide whether there is a better way to handle required info
        if errors:
            sub_errors = errors.errors()
            # If insert is not true we cannot check required fields as we are inserting with primary key
            if not insert:
                sub_errors = list(filter(lambda x: x.get('type') != 'value_error.missing', sub_errors))
            # if we still have sub-errors raise the errors to the user
            if sub_errors:
                raise errors
        base_fields = {'id', 'etl_step_id', 'created_at'}
        return {
            k: v
            for k, v in validated_data.items()
            if k not in base_fields and (use_defaults or k in fields_set)
        }

    def _strict_validate(self, input_data: Dict[str, Any], use_defaults: bool = False, insert: bool = False):
        # Check basic validation
        self._validate(input_data, use_defaults=use_defaults, insert=insert)
        # Add additional check for expected type for strict typing
        errors = []
        for key, val in input_data.items():
            type_str = self.attributes[key]
            # if the type str ends with brackets the expected type is an array
            # TODO check the elements types for lists
            if type_str.endswith('[]'):
                expected_type: type = list
            else:
                expected_type = column_registry[type_str].get_python_type()
            # If the val is not the expected type append a TypeError and continue on so we can collect all the errors
            if not isinstance(val, expected_type):
                errors.append(
                    ErrorWrapper(
                        TypeError(
                            "Strict validation found an error:\n"
                            f"{key!r} on Entity {self.name!r} has column type of {type_str!r} which expects a python type of {expected_type!r}, but provided value was type {type(val)}"
                        ),
                        loc=key,
                    )
                )
        # If we found any errors raise them
        if errors:
            raise PydValidationError(errors, self.__class__)
        return input_data

    def _get_hash(self, arg_dict: Dict[str, Any]) -> UUID:
        id_fks = (str(arg_dict[val]) for val in sorted(self.identifying_foreign_keys))
        id_attrs = []
        for attr_name in sorted(self.identifying_attributes):
            type_str = self.attributes[attr_name]
            data_type = column_registry[type_str]
            type_func = list if type_str == data_type.array_oid else data_type.python_type
            try:
                arg_val = arg_dict[attr_name]
                if arg_val is None or isinstance(arg_val, type_func):
                    coerced_val = arg_val
                else:
                    raise TypeError(
                        f"Type Coercing is turned off. You are trying to insert into attribute {self.name}({attr_name}) which has a type of {type_func} but you provided a type {type(arg_val)}.\n"
                        "If you want to turn Type Coercement on set the configuration variable DBGEN_TYPE_COERCING=true in your config file or environment variable"
                    )
            except TypeError as exc:
                raise ValueError(f"Error coercing value {arg_val!r} to type {type_func}:\n{exc}") from exc
            except KeyError:
                raise KeyError(f"Cannot find id_attribute {attr_name!r} in arg_dict for hashing: {arg_dict}")
            id_attrs.append(coerced_val)
        tuple_to_hash = (*id_attrs, *id_fks)
        return hash_tuple(tuple_to_hash)

    @property
    def full_name(self) -> str:
        return f"{self.schema_}.{self.name}" if self.schema_ else self.name

    @property
    def identifiers(self) -> Set[str]:
        return self.identifying_foreign_keys.union(self.identifying_attributes)


T = TypeVar('T')


class Load(ComputationalNode[T]):
    load_entity: LoadEntity
    primary_key: Optional[Union[Arg, Constant]] = None
    _output: Dict[UUID, Sequence[Any]] = PrivateAttr(default_factory=dict)
    insert: bool = False
    validation: Optional[ValidationEnum] = None
    outputs: List[str] = Field(default_factory=list)
    # _logger_name: ClassVar[
    #     Callable[["Base", Dict[str, Any]], str]
    # ] = lambda _, kwargs: f"dbgen.load.{kwargs.get('load_entity').name}"  # type: ignore

    @root_validator
    def change_output_name(cls, values):
        load_entity = values.get('load_entity')
        values['outputs'] = [f"{load_entity.name}_id"]
        return values

    @root_validator
    def check_bad_insert(cls, values):
        primary_key = values.get("primary_key")
        insert = values.get("insert")
        if primary_key is not None and insert:
            raise ValueError(
                f"Can't insert into {values.get('load_entity')} if we already have Primary Key..."
            )
        return values

    @validator("primary_key")
    def check_const_primary_key(cls, primary_key):
        if isinstance(primary_key, Constant):
            assert (
                primary_key.val is None
            ), f"Currently don't allow const primary keys to be used unless None is the value: {primary_key}"
        return primary_key

    @root_validator
    def check_required_info(cls, values):
        insert = values.get("insert")
        load_entity = values.get("load_entity")
        inputs = values.get("inputs")
        pk = values.get("primary_key")
        if load_entity is None:
            raise ValueError("Missing load_entity")
        pk_is_missing = pk is None or (isinstance(pk, Constant) and pk.val is None)
        if insert or pk_is_missing:
            missing_attrs = filter(lambda x: x not in inputs, load_entity.identifying_attributes)
            missing_fks = filter(lambda x: x not in inputs, load_entity.identifying_foreign_keys)
            missing = list(chain(missing_attrs, missing_fks))
            if missing:
                action = "insert" if insert else "update"
                raise ValueError(f"Can't {action} {load_entity} row without required ID info {missing}: {pk}")
        return values

    def __str__(self):
        return f"Load<{self.load_entity.name}, {len(self.inputs)} attrs/fks>"

    def _get_dependency(self) -> Dependency:
        tables_yielded: Set[str] = set()
        tables_needed: Set[str] = set()
        if self.insert:
            tables_yielded.add(self.load_entity.full_name)
        else:
            tables_needed.add(self.load_entity.full_name)

        columns_yielded: Set[str] = set()
        for attr in self.inputs:
            if self.insert or (attr not in self.load_entity.identifiers):
                columns_yielded.add(f"{self.load_entity.full_name}.{attr}")

        if not self.insert:
            columns_needed = {
                f"{self.load_entity.full_name}.{attr}"
                for attr in self.inputs
                if attr in self.load_entity.identifiers
            }
        else:
            columns_needed = set()

        return Dependency(
            tables_yielded=tables_yielded,
            columns_yielded=columns_yielded,
            columns_needed=columns_needed,
            tables_needed=tables_needed,
        )

    def new_run(
        self, row: Dict[str, Mapping[str, Any]], rows_to_load: Dict[str, Dict[UUID, Any]]
    ) -> Dict[str, List[UUID]]:
        not_list = lambda x: not isinstance(x, (list, tuple))
        lists_allowed = lambda x: self.load_entity.attributes[x].endswith('[]')
        is_list_of_lists = lambda x: isinstance(x, list) and x and isinstance(x[0], list)
        arg_dict = {
            key: [val] if (not_list(val) or (lists_allowed(key) and not is_list_of_lists)) else val
            for key, val in sorted(self._get_inputs(row).items())
        }
        # Check for empty lists, as that will cause the row to be ignored
        if any(map(lambda x: len(x) == 0, arg_dict.values())):
            # self._logger.debug(f'Row {arg_dict} produced 0 rows for load {self}')
            return {self.outputs[0]: []}
        # Check for broadcastability
        try:
            is_broadcastable(*arg_dict.values())
        except ValueError as exc:
            raise ValueError(
                f"While assembling rows for loading into {self} found two sequences in the inputs to the load with non-equal, >1 length\n"
                "This can occur when two outputs from pyblocks/queries are unequal in length.\n"
                "If so, please make the relevant cartesian product of the two lists before loading\n"
            ) from exc
        try:
            broadcasted_values = [
                {key: val for key, val in zip(arg_dict.keys(), row)} for row in broadcast(*arg_dict.values())
            ]
        except ValueError as exc:
            raise ValueError(
                f"Error occurred during the loading of row {(str(arg_dict)[:100])} into entity {self.load_entity.full_name}"
            ) from exc

        # Load
        validation = self.validation or config.validation
        if validation == ValidationEnum.STRICT:
            validation_func = self.load_entity._strict_validate
        elif validation == ValidationEnum.COERCE:
            validation_func = self.load_entity._validate

        try:
            broadcasted_values = list(map(partial(validation_func, insert=self.insert), broadcasted_values))
        except PydValidationError as exc:
            raise ValidationError(
                f"Error occurred during data validation while loading into {self.load_entity.full_name!r}:\n{exc}"
            ) from exc
        # If we have a user supplied Primary Key go get it and broadcast it
        if self.primary_key is not None:
            primary_arg_val = self.primary_key.arg_get(row)
            # Validate the primary key type
            if isinstance(primary_arg_val, UUID):
                primary_keys: List[UUID] = [primary_arg_val]
            elif isinstance(primary_arg_val, Sequence):
                assert all(map(lambda x: isinstance(x, UUID), primary_arg_val))
                primary_keys = list(primary_arg_val)
            else:
                ValueError(f"Unknown Primary Key Type: {primary_arg_val}")

            if len(primary_keys) == 1:
                primary_keys *= len(broadcasted_values)
            elif len(primary_keys) != len(broadcasted_values):
                raise ValueError(
                    f"Cannot broadcast Primary Key to Max Length: {len(primary_keys)} {len(broadcasted_values)}"
                )
        else:
            # If we don't have a primary key get it from the identifying info on the broadcasted
            # Values
            primary_keys = [self.load_entity._get_hash(row) for row in broadcasted_values]
        sorted_keys = sorted(self.inputs.keys())
        # Update rows to load dict in place
        rows_to_load[self.hash].update(
            {
                primary_key: [value[key] for key in sorted_keys]
                for primary_key, value in zip(primary_keys, broadcasted_values)
            }
        )

        return {self.outputs[0]: primary_keys}

    def _load_data(self, data: Mapping[UUID, tuple], connection: 'Connection', etl_step_id: UUID):
        """Run the Load statement for the given namespace rows.

        Args:
            namespace_rows (List[Dict[str, Any]]): A dictionary with the strings as hashes and the values as the local namespace dictionaries from PyBlocks, Queries, and Consts
        """
        self._logger.debug(f"Loading into {self.load_entity.name}")
        load_data(
            data,
            connection,
            self.load_entity,
            self.inputs.keys(),
            self.insert,
            self.load_entity.hash,
            etl_step_id,
        )

    async def _async_load(self, data, connection: 'AsyncConnection', etl_step_id: UUID) -> int:
        return await async_load_data(
            data,
            connection,
            self.load_entity,
            self.inputs.keys(),
            self.insert,
            self.load_entity.hash,
            etl_step_id,
        )



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Objects related to the running of Models and ETLSteps."""
from bdb import BdbQuit
from math import ceil
from time import time
from traceback import format_exc
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple
from uuid import UUID

from psycopg import connect as pg3_connect
from pydasher import hasher
from sqlalchemy.future import Engine
from sqlmodel import Session, select

import dbgen.exceptions as exceptions
from dbgen.configuration import config
from dbgen.core.base import Base, encoders
from dbgen.core.dashboard import BarNames, Dashboard
from dbgen.core.etl_step import ETLStep
from dbgen.core.metadata import ETLStepEntity, ETLStepRunEntity, Repeats, RunEntity, Status
from dbgen.core.node.extract import Extract
from dbgen.core.node.query import BaseQuery, ExternalQuery
from dbgen.core.run.async_run import AsyncETLStepExecutor
from dbgen.core.run.utilities import BaseETLStepExecutor, RunConfig, update_run_by_id
from dbgen.exceptions import SerializationError
from dbgen.utils.typing import NAMESPACE_TYPE

if TYPE_CHECKING:
    from psycopg import Connection as PG3Connection


class ETLStepExecutor(BaseETLStepExecutor):
    """Synchronous ETLStep Executor."""

    def execute(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        meta_session: Session,
        run_id: Optional[int],
        dashboard: Optional[Dashboard],
        etl_step_run: ETLStepRunEntity,
    ) -> int:
        # Start the ETLStep
        self._logger.info(f'Running ETLStep {self.etl_step.name!r}...')
        self._etl_step_run = etl_step_run
        self._etl_step_run.status = Status.running
        meta_session.commit()
        start = time()
        self._logger.debug('Fetching repeats')
        # Query the repeats table for input_hashes that match this etl_step's hash
        self._old_repeats = set(
            meta_session.exec(
                select(Repeats.input_hash).where(Repeats.etl_step_id == self.etl_step.uuid)
            ).all()
        )

        # Setup the extractor
        self._logger.debug('Initializing extractor')
        extract = self.etl_step.extract
        extract._set_run_config(self.run_config)
        with main_engine.connect() as extractor_connection:
            try:
                with extract:
                    # Specifically handle Query extracts by passing in the connection to the database to key methods
                    if isinstance(extract, BaseQuery) and not isinstance(extract, ExternalQuery):
                        extract.set_connection(
                            connection=extractor_connection, yield_per=self.etl_step.batch_size
                        )

                    self._logger.debug('Fetching extractor length')
                    row_count = extract.length() if not self.run_config.skip_row_count else None
                    # Commit the row count to the metadatabase
                    self._etl_step_run.inputs_extracted = row_count
                    meta_session.commit()
                    # The batch_size is set either on the run_config or the etl_step
                    batch_size = self.run_config.batch_size or self.etl_step.batch_size
                    if batch_size is None and row_count:
                        batch_size = ceil(row_count / self.run_config.batch_number)
                    elif batch_size is None:
                        batch_size = config.batch_size
                    # Check for invalid batch sizess
                    if batch_size is not None and batch_size < 0:
                        raise ValueError(f"Invalid batch size batch_size must be >0: {batch_size}")

                    # Open raw connections for fast loading
                    main_raw_connection = pg3_connect(str(main_engine.url))
                    meta_raw_connection = pg3_connect(str(meta_engine.url))
                    # Start while loop to iterate through the nodes
                    self._logger.debug('Looping through extracted rows...')
                    if dashboard is not None:
                        dashboard.add_etl_progress_bars(total=row_count)
                    for batch_ind, batch in enumerate(self.batchify(extract, batch_size, dashboard)):
                        (
                            _,
                            rows_to_load,
                            rows_processed,
                            inputs_skipped,
                            exc,
                        ) = self.etl_step.transform_batch(batch, self.run_config)
                        # Check if transforms or loads raised an error
                        if exc:
                            msg = f"Error when running etl_step {self.etl_step.name}"
                            self._logger.error(msg)
                            self._etl_step_run.status = Status.failed
                            self._etl_step_run.error = exc
                            run = meta_session.get(RunEntity, run_id)
                            assert run
                            run.errors = run.errors + 1 if run.errors else 1
                            meta_session.commit()
                            meta_session.close()
                            return 1
                        if dashboard is not None:
                            dashboard.advance_bar(BarNames.TRANSFORMED, advance=len(batch))
                        rows_inserted, rows_updated = self._load_data(
                            rows_to_load, connection=main_raw_connection
                        )
                        if dashboard is not None:
                            dashboard.advance_bar(BarNames.LOADED, advance=rows_processed)
                        self._load_repeats(meta_raw_connection)
                        self._logger.debug(
                            f'Done loading batch {batch_ind}. Inserted {rows_inserted} and updated {rows_updated} rows.'
                        )

                        # Commit changes to db
                        self._etl_step_run.rows_inserted += rows_inserted
                        self._etl_step_run.rows_updated += rows_updated
                        self._etl_step_run.inputs_skipped += inputs_skipped
                        meta_session.commit()

                # Finish the run and commit to DB
                self._etl_step_run.status = Status.completed
                self._etl_step_run.runtime = round(time() - start, 3)
                self._logger.info(
                    f"Finished running etl_step {self.etl_step.name} in {self._etl_step_run.runtime}(s)."
                )
                self._logger.info(f"Loaded approximately {self._etl_step_run.rows_inserted} rows")
                meta_session.commit()
                meta_session.close()
                return 0
            except (Exception, KeyboardInterrupt, SystemExit, BdbQuit) as exc:
                # Need to catch User Cancels and cancel the whole run
                update_run_by_id(run_id, Status.failed, meta_session)
                if isinstance(exc, (KeyboardInterrupt, BdbQuit)):
                    self._logger.info('Shutting down due to keyboardInterrupt')
                else:
                    msg = f"Uncaught exception while running etl_step {self.etl_step.name}"
                    self._logger.error(msg)
                self._etl_step_run.status = Status.failed
                self._etl_step_run.error = format_exc(limit=1)
                run = meta_session.get(RunEntity, run_id)
                assert run
                run.errors = run.errors + 1 if run.errors else 1
                meta_session.commit()
                meta_session.close()
                raise

    def batchify(
        self, extract: Extract, batch_size: int, dashboard: Optional[Dashboard]
    ) -> Generator[List[Tuple[UUID, NAMESPACE_TYPE]], None, None]:
        # initialize the batch and counts
        batch: List[Tuple[UUID, NAMESPACE_TYPE]] = []
        self._etl_step_run.inputs_extracted = 0
        self._etl_step_run.unique_inputs = 0
        # Loop the the rows in the extract function
        for row in extract.extract():
            # Check the hash of the inputs against the metadatabase
            processed_row = extract.process_row(row)
            is_repeat, input_hash = self._check_repeat(processed_row, self.etl_step.uuid)

            # If we are running with --retry redo repeats
            if self.run_config.retry or not is_repeat:
                # Store the hash for newly seen rows for later loading
                self._etl_step_run.inputs_extracted += 1
                if not is_repeat:
                    self._etl_step_run.unique_inputs += 1
                    self._new_repeats.add(input_hash)
                batch.append((input_hash, {extract.hash: processed_row}))
            elif dashboard is not None:
                dashboard.advance_bar(BarNames.EXTRACTED, advance=1)

            # If batch size is reached advance bar and yield
            if len(batch) == batch_size:
                if dashboard is not None:
                    dashboard.advance_bar(BarNames.EXTRACTED, advance=len(batch))
                yield batch
                batch = []

        # Load the remaining rows
        if batch:
            if dashboard is not None:
                dashboard.advance_bar(BarNames.EXTRACTED, advance=len(batch))
            yield batch
        if dashboard is not None:
            dashboard.set_total(total=self._etl_step_run.inputs_extracted)

    def _load_data(self, rows_to_load: Dict[str, Dict[UUID, Any]], connection) -> Tuple[int, int]:
        rows_inserted = 0
        rows_updated = 0
        for load in self.etl_step._sorted_loads():
            self._logger.debug(f'Loading into {load}')
            rows = rows_to_load[load.hash]
            if load.insert:
                rows_inserted += len(rows)
            else:
                rows_updated += len(rows)
            load._load_data(data=rows, connection=connection, etl_step_id=self.etl_step.uuid)
        return (rows_inserted, rows_updated)

    def _load_repeats(self, connection: 'PG3Connection') -> None:
        rows = {input_hash: (self.etl_step.uuid,) for input_hash in self._new_repeats}
        Repeats._quick_load(connection, rows, column_names=["etl_step_id"])
        self._old_repeats = self._old_repeats.union(self._new_repeats)
        self._new_repeats = set()

    def _check_repeat(self, extracted_dict: Dict[str, Any], etl_step_uuid: UUID) -> Tuple[bool, UUID]:
        # Convert Row to a dictionary so we can hash it for repeat-checking
        input_hash = UUID(hasher((etl_step_uuid, extracted_dict), encoders=encoders))
        # If the input_hash has been seen and we don't have retry=True skip row
        is_repeat = input_hash in self._old_repeats or input_hash in self._new_repeats
        return (is_repeat, input_hash)


class BaseETLStepRun(Base):
    """A lightweight wrapper for the ETLStep that grabs a specific ETLStep from metadatabase and runs it."""

    _run_config: RunConfig = RunConfig()

    class Config:
        """Pydantic COnfig"""

        underscore_attrs_are_private = True

    def get_etl_step(self, meta_engine: Engine, *args, **kwargs) -> ETLStep:
        raise NotImplementedError

    def get_executor(self, etl_step, run_config) -> BaseETLStepExecutor:
        raise NotImplementedError

    def execute(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        run_id: Optional[int],
        run_config: Optional[RunConfig],
        ordering: Optional[int],
        dashboard: Optional[Dashboard] = None,
    ):
        # Set default values for run_config if none provided
        if run_config:
            self._run_config = run_config

        etl_step = self.get_etl_step(meta_engine=meta_engine)
        # Initialize the etl_step_row in the meta database
        meta_session = Session(meta_engine)
        etl_step_run = self._initialize_etl_step_run(
            etl_step=etl_step, session=meta_session, run_id=run_id, ordering=ordering
        )
        # Check if our run config excludes our etl_step
        if not self._run_config.should_etl_step_run(etl_step):
            self._logger.info(f'Excluding etl_step {etl_step.name!r}')
            etl_step_run.status = Status.excluded
            meta_session.commit()
            return
        elif etl_step.name in self._run_config.upstream_fail_exclude:
            self._logger.info(f'Excluding etl_step {etl_step.name!r} due to upstream failure')
            etl_step_run.status = Status.upstream_failed
            meta_session.commit()
            return
        # Initialize the ETLStepExecutor
        executor = self.get_executor(etl_step, self._run_config)
        return_code = executor.execute(
            main_engine=main_engine,
            meta_engine=meta_engine,
            meta_session=meta_session,
            run_id=run_id,
            dashboard=dashboard,
            etl_step_run=etl_step_run,
        )
        return return_code

    def _initialize_etl_step_run(
        self,
        session: Session,
        etl_step: ETLStep,
        run_id: Optional[int],
        ordering: Optional[int],
    ) -> ETLStepRunEntity:
        # if no run_id is provided create one and mark it as a testing run
        if run_id is None:
            run = RunEntity(status='testing')
            session.add(run)
            session.commit()
            session.refresh(run)
            ordering = 0
            run_id = run.id
        etl_step_row = etl_step._get_etl_step_row()
        session.merge(etl_step_row)
        session.commit()
        query = etl_step.extract.render_query() if isinstance(etl_step.extract, BaseQuery) else ''
        etl_step_run = ETLStepRunEntity(
            run_id=run_id,
            etl_step_id=etl_step_row.id,
            status=Status.initialized,
            ordering=ordering,
            query=query,
        )
        session.add(etl_step_run)
        session.commit()
        session.refresh(etl_step_run)
        return etl_step_run


class ETLStepRun(BaseETLStepRun):
    etl_step: ETLStep

    def get_etl_step(self, meta_engine: Engine, *args, **kwargs):
        return self.etl_step

    def get_executor(self, etl_step, run_config) -> BaseETLStepExecutor:
        return ETLStepExecutor(etl_step=etl_step, run_config=run_config)


class RemoteETLStepRun(BaseETLStepRun):
    etl_step_id: UUID

    def get_etl_step(self, meta_engine, *args, **kwargs):
        with Session(meta_engine) as sess:
            etl_step_json = sess.exec(
                select(ETLStepEntity.etl_step_json).where(ETLStepEntity.id == self.etl_step_id)
            ).one()
            try:
                etl_step = ETLStep.deserialize(etl_step_json)
            except ModuleNotFoundError as exc:
                import os

                raise SerializationError(
                    f"While deserializing etl_step id {self.etl_step_id} an unknown module was encountered. Are you using custom dbgen objects reachable by your python environment? Make sure any custom extractors or code can be found in your PYTHONPATH environment variable\nError: {exc}\nPYTHONPATH={os.environ.get('PYTHONPATH')}"
                ) from exc
            if etl_step.uuid != self.etl_step_id:
                error = f"Deserialization Failed the etl_step hash has changed for etl_step named {etl_step.name}!\n{etl_step}\n{self.etl_step_id}"
                raise exceptions.SerializationError(error)
        return etl_step

    def get_executor(self, etl_step, run_config) -> BaseETLStepExecutor:
        return ETLStepExecutor(etl_step=etl_step, run_config=run_config)


class AsyncETLStepRun(BaseETLStepRun):
    etl_step: ETLStep

    def get_etl_step(self, meta_engine: Engine, *args, **kwargs):
        return self.etl_step

    def get_executor(self, etl_step, run_config) -> BaseETLStepExecutor:
        return AsyncETLStepExecutor(etl_step=etl_step, run_config=run_config)


class AsyncRemoteETLStepRun(BaseETLStepRun):
    etl_step_id: UUID

    def get_etl_step(self, meta_engine, *args, **kwargs):
        with Session(meta_engine) as sess:
            etl_step_json = sess.exec(
                select(ETLStepEntity.etl_step_json).where(ETLStepEntity.id == self.etl_step_id)
            ).one()
            try:
                etl_step = ETLStep.deserialize(etl_step_json)
            except ModuleNotFoundError as exc:
                import os

                raise SerializationError(
                    f"While deserializing etl_step id {self.etl_step_id} an unknown module was encountered. Are you using custom dbgen objects reachable by your python environment? Make sure any custom extractors or code can be found in your PYTHONPATH environment variable\nError: {exc}\nPYTHONPATH={os.environ.get('PYTHONPATH')}"
                ) from exc
            if etl_step.uuid != self.etl_step_id:
                error = f"Deserialization Failed the etl_step hash has changed for etl_step named {etl_step.name}!\n{etl_step}\n{self.etl_step_id}"
                raise exceptions.SerializationError(error)
        return etl_step

    def get_executor(self, etl_step, run_config) -> BaseETLStepExecutor:
        return AsyncETLStepExecutor(etl_step=etl_step, run_config=run_config)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Exceptions used by DBgen"""


class DBgenException(Exception):
    """
    Base class for all DBgen's errors.
    Each custom exception should be derived from this class
    """

    def __init__(self, msg="<No Message>"):
        self.msg = msg


class DBgenExternalError(DBgenException):
    """Custom class for catching errors that occur in code external to dbgen"""

    status_code = 400


class DBgenInternalError(DBgenException):
    """
    Custom class for catching errors that occur in code internal to dbgen
    """


class DBgenETLStepError(DBgenException):
    """Custom class for catching errors that occur in code external to dbgen"""


class DBgenSkipException(DBgenException):
    """Custom class for errors raised when a ETLStep is skipped"""


class DBgenNotFoundException(DBgenException):
    """Raise when the requested object/resource is not available in the system"""

    status_code = 404


class DBgenConfigException(DBgenException):
    """Raise when there is configuration problem"""


class DBgenLoadException(DBgenException):
    """Raise when there is loading problem"""


class DBgenPythonTransformError(DBgenException):
    """Raise when there is loading problem"""


class DBgenMissingInfo(DBgenException):
    """Raise when there is missing info in a class/function initialization"""


class InvalidArgument(DBgenException):
    """Raise when there is missing info in a class/function initialization"""


class NodeUsedAsInput(InvalidArgument):
    """Raise when a PythonTransform is used as input to ComputationalNode"""


class DBgenTypeError(DBgenException):
    """Raise when a value cannot be cast the correct type"""


class QueryParsingError(DBgenException):
    """Error raised when parsing a sqlalchemy select statement."""


class SerializationError(DBgenException):
    """Error raised when serializing and deserializing dbgen objects."""


class RepeatException(DBgenException):
    """Error raised when serializing and deserializing dbgen objects."""


class DatabaseError(DBgenException):
    """Error raised when interacting with the database"""


class ValidationError(DBgenException):
    """Error raised when data fails validation"""


class AsyncRunError(DBgenException):
    """Error raised during async run"""


class ExtractorError(AsyncRunError):
    """Error raised by the async extractor routine"""


class TransformerError(AsyncRunError):
    """Error raised by the async extractor routine"""


class ResultsError(AsyncRunError):
    """Error raised by the async extractor routine"""


class LoaderError(AsyncRunError):
    """Error raised by the async extractor routine"""


class ModelRunError(DBgenException):
    """Error raised when a model is run."""


class MissingImportError(DBgenException):
    """Error raised an import is missing."""



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""For parsing in config YAML."""
import os
import tempfile
from enum import Enum
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Optional, Tuple

from pydantic import BaseSettings, PostgresDsn, SecretStr, validator
from pydantic.tools import parse_obj_as

from dbgen.utils.log import LogLevel, setup_logger
from dbgen.utils.sql import Connection

if TYPE_CHECKING:
    from sqlalchemy.future import Engine  # pragma: no cover


# Force postgresql schemes for connection for sqlalchemy
class PostgresqlDsn(PostgresDsn):
    allowed_schemes = {"postgresql"}
    path: str


class ValidationEnum(str, Enum):
    COERCE = 'coerce'
    STRICT = 'strict'
    OFF = 'off'


hidden_options = ('pdb', 'testing')


class DBgenConfiguration(BaseSettings):
    """Settings for the pg4j, especially database connections."""

    main_dsn: PostgresqlDsn = parse_obj_as(PostgresqlDsn, "postgresql://postgres@localhost:5432/dbgen")
    main_schema: str = "public"
    main_password: SecretStr = ""  # type: ignore
    meta_dsn: Optional[PostgresqlDsn]
    meta_schema: str = "dbgen_log"
    meta_password: Optional[SecretStr] = None
    model_str: str = 'model.main:make_model'
    batch_size: int = 100
    temp_dir: Path = Path(tempfile.gettempdir())
    validation: ValidationEnum = ValidationEnum.COERCE
    pdb: bool = False
    testing: bool = False
    log_level: LogLevel = LogLevel.INFO

    class Config:
        """Pydantic configuration"""

        env_file = os.environ.get("DBGEN_CONFIG", ".env")
        env_prefix = "DBGEN_"
        extra = "forbid"

    @validator('temp_dir')
    def validate_temp_dir(cls, temp_dir: Path) -> Path:
        if not temp_dir.exists():
            temp_dir.mkdir(exist_ok=True, parents=True)
        if temp_dir.is_dir():
            return temp_dir
        raise ValueError(f"temp_dir value {temp_dir} is not a directory")

    @validator('batch_size')
    def validate_batch_size(cls, batch_size) -> Path:
        if batch_size > 1:
            return batch_size
        raise ValueError(f'batch_size must be a positive integer: {batch_size}')

    def display(self, show_defaults: bool = False, show_passwords: bool = False):
        params = []
        for key, val in self.dict().items():
            if val is not None and key not in hidden_options:
                str_val = f'{val.get_secret_value()}' if show_passwords and "password" in key else val
                if show_defaults or key in self.__fields_set__:
                    params.append(f"dbgen_{key} = {str_val}")
                else:
                    params.append(f"# dbgen_{key} = {str_val}")

        params_str = "\n".join(params)
        output = f"""# DBgen Settings\n{params_str}"""
        return dedent(output)

    def update(self, new_config: 'DBgenConfiguration', set_defaults: bool = False):
        for field in self.__fields__:
            if field in new_config.__fields_set__ or set_defaults:
                setattr(self, field, getattr(new_config, field))

    def __str__(self):
        return self.display()

    def __repr__(self):
        return self.display()


config = DBgenConfiguration()
root_logger, stdout_handler = setup_logger(config.log_level, config.log_level)


def update_config(config_file: 'Path') -> DBgenConfiguration:
    global config
    input_config = DBgenConfiguration(_env_file=config_file)
    if input_config.meta_schema != config.meta_schema:
        root_logger.warning(
            f"WARNING: Input config file {str(config_file)!r} has a meta_schema value set that is different than the current value at {DBgenConfiguration.Config.env_file!r}."
            " This setting will be ignored as setting the meta_schema in an CLI input config file is not fully implemented."
            " To set this variable, please use the DBGEN_CONFIG environmental variable rather than the -c/--config CLI argument."
            f" For example, set `DBGEN_CONFIG={str(config_file)!r}` before running dbgen."
        )
    config.update(input_config)
    return config


def get_connections() -> Tuple['Connection', 'Connection']:
    main_conn = Connection.from_uri(config.main_dsn, config.main_schema, password=config.main_password)
    meta_dsn = config.meta_dsn or config.main_dsn
    meta_password = config.meta_password or config.main_password
    meta_conn = Connection.from_uri(meta_dsn, config.meta_schema, meta_password)
    return (main_conn, meta_conn)


def get_engines(echo: bool = False) -> Tuple['Engine', 'Engine']:
    (main_conn, meta_conn) = get_connections()
    return (main_conn.get_engine(echo), meta_conn.get_engine(echo))



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from itertools import chain
from json import dumps
from pathlib import Path
from typing import List
from uuid import UUID

import typer
from rich.console import Console
from rich.table import Table
from sqlalchemy import update
from sqlmodel import Session, select

import dbgen.cli.styles as styles
from dbgen.cli.options import (
    chdir_option,
    config_option,
    model_arg_option,
    model_string_option,
    verbose_option,
)
from dbgen.cli.utils import state, test_connection, validate_model_str
from dbgen.configuration import config, get_connections
from dbgen.core.metadata import ModelEntity

model_app = typer.Typer(name='model', no_args_is_help=True)


@model_app.command('list')
def list_models(config_file: Path = config_option, tags: List[str] = typer.Option(None, '-t')):
    """List the models in the metadatabase."""
    # Notify of config file
    _, meta_conn = get_connections()
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()
    tags = tags or []
    statement = select(  # type: ignore
        ModelEntity.id,
        ModelEntity.name,
        ModelEntity.created_at,
        ModelEntity.last_run,
        ModelEntity.tags,
    )  # type: ignore
    if tags:
        statement = statement.where(ModelEntity.tags.op('&&')(tags))  # type: ignore
    columns = ['id', 'name', 'created_at', 'last_run', 'tags']
    table = Table(
        *columns,
        show_lines=True,
        highlight=True,
        border_style='magenta',
    )
    with Session(meta_engine) as session:
        result = session.exec(statement)
        for model_id, model_name, created_at, last_run, tags in result:
            table.add_row(*map(str, (model_id, model_name, created_at, last_run, tags)))
    console = Console()
    console.print(table)


@model_app.command('tag')
def tag(model_id: UUID, tags: List[str], config_file: Path = config_option):
    # Notify of config file
    _, meta_conn = get_connections()
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()

    with Session(meta_engine) as session:
        existing_tags = session.exec(select(ModelEntity.tags).where(ModelEntity.id == model_id)).one_or_none()
        if existing_tags is None:
            raise typer.BadParameter(f"Invalid model_id, no model with ID {model_id}")
        new_tags = set(chain(existing_tags, tags))
        session.execute(update(ModelEntity).values(tags=new_tags).where(ModelEntity.id == model_id))

        session.commit()


@model_app.command('serialize')
def model_serialize(
    model_str: str = model_arg_option,
    out_file: Path = typer.Option(
        None, '-o', '--out', help='Path to write the serialized model to in json format'
    ),
    config_file: Path = config_option,
):
    model = validate_model_str(model_str)

    # Notify of config file
    _, meta_conn = get_connections()
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()
    with Session(meta_engine) as session:
        model_row = model._get_model_row()
        # Check for existing row and if found grab its created_at
        created_at = session.exec(
            select(ModelEntity.created_at).where(ModelEntity.id == model.uuid)
        ).one_or_none()
        if created_at is None:
            session.merge(model_row)
            session.commit()
            styles.good_typer_print(f"Loaded model {model.name!r} into the database with ID {model.uuid}")
        else:
            styles.good_typer_print(f"Model {model.name!r} already existed with ID {model.uuid}")
    if out_file:
        out_file.write_text(dumps(model_row.graph_json))
        styles.good_typer_print(f"Wrote serialized graph to {out_file}")


@model_app.command('export')
def model_export(
    model_id: UUID,
    out_file: Path = typer.Option(
        'model.json', '-o', '--out', help='Path to write the serialized model to in json format'
    ),
    config_file: Path = config_option,
):

    # Notify of config file
    _, meta_conn = get_connections()
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()

    with Session(meta_engine) as session:
        # Check for existing row and if found grab its created_at
        graph_json = session.exec(
            select(ModelEntity.graph_json).where(ModelEntity.id == model_id)
        ).one_or_none()
        if not graph_json:
            raise ValueError(f"Invalid model_id: No model found with model_id {model_id}")

    out_file.write_text(dumps(graph_json))
    styles.good_typer_print(f"Wrote serialized graph to {out_file}")


@model_app.command('validate')
def validate(
    model_str: str = model_string_option,
    config_file: Path = config_option,
    _verbose: bool = verbose_option(),
    _chdir: Path = chdir_option,
):
    """Quick utility method for quickly validating a model will compile without any need for database connections."""
    # Start connection from config
    # Use config model_str if none is provided
    model_str = model_str or config.model_str
    model = validate_model_str(model_str)
    styles.good_typer_print(f"Model(name={model.name!r}) was successfully validated")
    styles.good_typer_print(
        f'Model contains {len(model.etl_steps)} ETLStep(s) and {len(model.registry.metadata.tables)} two entities.'
    )
    if state['verbose']:
        styles.good_typer_print(f'Model contains {len(model.etl_steps)} ETLStep(s).')



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import csv
import re
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, TextIO, Tuple, TypeVar

import yaml
from pydantic import DirectoryPath, root_validator

from dbgen.core.node.extract import Extract

T = TypeVar('T')
T2 = TypeVar('T2', bound=str)


class _FileExtractorBase(Extract[T]):
    directory: DirectoryPath
    pattern: Optional[str]
    recursive: bool = False
    _file_paths: List[Path]

    def setup(self, **_):
        file_gen = self.directory.rglob('*') if self.recursive else self.directory.iterdir()
        self._file_paths = [
            x for x in file_gen if x.is_file() and (self.pattern is None or re.search(self.pattern, x.name))
        ]

    def length(self, **_):
        return len(self._file_paths)


class FileNameExtractor(_FileExtractorBase[Path]):
    outputs: List[str] = ['file_name']

    def extract(self) -> Generator[Path, None, None]:
        yield from self._file_paths


class FileExtractor(_FileExtractorBase[Tuple[Path, Any]]):
    outputs: List[str] = ['file_name', 'parsed_file']

    def file_parser(self, file_obj: TextIO) -> str:
        return file_obj.read()

    def extract(self) -> Generator[Tuple[Path, Any], None, None]:
        for file_name in self._file_paths:
            with open(file_name) as f:
                yield file_name, self.file_parser(f)


class YamlExtractor(FileExtractor):
    pattern: Optional[str] = r'.*\.(yaml|yml)'

    def file_parser(self, file_obj: TextIO):
        return yaml.safe_load(file_obj)


class CSVExtractor(Extract[Dict[str, str]]):
    """Extract each line in a CSV into its own input row."""

    path: Path
    ensure_path_exists: bool = True
    has_header: bool = True
    delimiter: str = ','
    outputs: List[str]
    _file: TextIOWrapper
    _reader: csv.DictReader

    @root_validator
    def validate_path_exists(cls, values):
        if not values['path'].exists() and values['ensure_path_exists']:
            raise ValueError(f"CSV {values['path']} does not exist.")
        return values

    def setup(self):
        self._file = self.path.open('r')
        fieldnames = None if self.has_header else self.outputs
        self._reader = csv.DictReader(self._file, fieldnames=fieldnames)

    def teardown(self):
        self._file.close()

    def length(self):
        count = sum(1 for _ in self._reader)
        self._file.seek(0)
        fieldnames = None if self.has_header else self.outputs
        self._reader = csv.DictReader(self._file, fieldnames=fieldnames)
        return count

    def extract(self) -> Generator[Dict[str, str], None, None]:
        yield from self._reader



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Utilities for converting and coercing types."""
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from json import dumps
from pathlib import PosixPath
from typing import Any
from uuid import UUID

from psycopg import adapters
from psycopg.adapt import Dumper
from psycopg.types.json import Jsonb, JsonbDumper, set_json_dumps


# Utilities for converting types
def escape_str_replace(text: Any) -> str:
    return str(text).replace('\\', '\\\\').replace('\n', '\\n').replace('\r', '\\r').replace('\t', '\\t')


datetime_types = (datetime, time, date)
datetime_to_str = lambda x: x.isoformat() if x is not None else str(x)
timedelta_to_str = lambda x: str(x.total_seconds()) if x is not None else str(x)
basic_to_str = lambda x: escape_str_replace(x)
bytes_to_str = lambda x: escape_str_replace(x.decode('utf-8')) if x is not None else str(x)

# TODO create central location for all json serialization (pydasher does hashing serialization)
def json_dumps(thing):
    def json_default(thing: Any) -> str:
        if isinstance(thing, datetime_types):
            return thing.isoformat()
        elif isinstance(thing, timedelta):
            return timedelta_to_str(thing)
        elif isinstance(thing, bytes):
            return bytes_to_str(thing)
        elif isinstance(thing, (Decimal, UUID)):
            return escape_str_replace(thing)
        elif isinstance(thing, set):
            return str(list(thing))
        raise TypeError(f"Cannot serialize object of type {type(thing)}: {thing}")

    return dumps(thing, default=json_default)


# Allow psycopg to dump non builtin types to database
set_json_dumps(json_dumps)


# Create a dumper for dict types
class DictDumper(Dumper):
    oid = adapters.types.get_oid('json')

    def dump(self, thing):
        return JsonbDumper(dict).dump(Jsonb(thing))


# Create a dumper for dict types
class PathDumper(Dumper):
    oid = adapters.types.get_oid('path')

    def dump(self, thing):
        return str(thing).encode(encoding='utf-8')


# Set the adapters for dicts and json
adapters.register_dumper(dict, DictDumper)
adapters.register_dumper(PosixPath, PathDumper)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from datetime import datetime, timedelta
from enum import Enum
from typing import List, Optional
from uuid import UUID

from sqlalchemy import Column, func
from sqlalchemy.orm import registry
from sqlalchemy.sql.expression import text
from sqlmodel import Field, Relationship, select
from sqlmodel.sql.sqltypes import GUID, AutoString

from dbgen.configuration import config
from dbgen.core.entity import BaseEntity, get_created_at_field, id_field
from dbgen.utils.sql import create_view

META_SCHEMA = config.meta_schema
meta_registry = registry()


class Status(str, Enum):
    initialized = "initialized"
    excluded = "excluded"
    running = "running"
    failed = "failed"
    completed = "completed"
    testing = "testing"
    upstream_failed = "upstream_failed"


class Root(BaseEntity):
    __schema__ = META_SCHEMA


class ModelETLStepMap(Root, registry=meta_registry, table=True):
    __tablename__ = 'model_etl_step'
    model_id: Optional[UUID] = Field(None, foreign_key=f"{META_SCHEMA}.model.id", primary_key=True)
    etl_step_id: Optional[UUID] = Field(None, foreign_key=f"{META_SCHEMA}.etl_step.id", primary_key=True)


class ModelEntity(Root, registry=meta_registry, table=True):
    __tablename__ = 'model'
    id: Optional[UUID] = id_field
    created_at: Optional[datetime] = get_created_at_field()
    last_run: Optional[datetime]
    name: str
    graph_json: Optional[dict]
    tags: List[str] = Field(default_factory=list)
    etl_steps: List['ETLStepEntity'] = Relationship(back_populates='models', link_model=ModelETLStepMap)


class ETLStepEntity(Root, registry=meta_registry, table=True):
    __tablename__ = "etl_step"
    id: Optional[UUID] = id_field
    name: str
    query: Optional[str]
    description: Optional[str]
    batch_size: Optional[int]
    tables_needed: Optional[str]
    table_yielded: Optional[str]
    column_needed: Optional[str]
    column_yielded: Optional[str]
    etl_step_json: Optional[dict]
    etl_step_runs: List['ETLStepRunEntity'] = Relationship(back_populates='etl_step')
    models: List['ModelEntity'] = Relationship(back_populates='etl_steps', link_model=ModelETLStepMap)


class RunEntity(Root, registry=meta_registry, table=True):
    __tablename__ = "run"
    id: Optional[int] = Field(None, sa_column_kwargs={"autoincrement": True, "primary_key": True})
    status: Optional[Status]
    build: Optional[bool]
    only: Optional[str]
    exclude: Optional[str]
    start: Optional[str]
    until: Optional[str]
    runtime: Optional[timedelta]
    errors: int = 0
    etl_step_runs: List['ETLStepRunEntity'] = Relationship(back_populates='run')


class ETLStepRunEntity(Root, registry=meta_registry, table=True):
    __tablename__ = "etl_step_run"
    etl_step_id: Optional[UUID] = ETLStepEntity.foreign_key(primary_key=True)
    run_id: Optional[int] = RunEntity.foreign_key(primary_key=True)
    created_at: Optional[datetime] = get_created_at_field()
    ordering: Optional[int]
    status: Optional[str]
    runtime: Optional[float]
    unique_inputs: int = 0
    inputs_skipped: int = 0
    inputs_extracted: Optional[int] = 0
    inputs_processed: int = 0
    rows_inserted: int = 0
    rows_updated: int = 0
    memory_usage: Optional[float]
    query: Optional[str]
    error: Optional[str]
    run: RunEntity = Relationship(back_populates='etl_step_runs')
    etl_step: ETLStepEntity = Relationship(back_populates='etl_step_runs')


class Repeats(Root, registry=meta_registry, table=True):
    __tablename__ = "repeats"
    etl_step_id: Optional[UUID] = ETLStepEntity.foreign_key()
    input_hash: UUID = Field(..., primary_key=True)


run_view_statement = (
    select(
        ETLStepEntity.name,
        *(getattr(ETLStepRunEntity, x) for x in ETLStepRunEntity.__fields__ if x not in ("etl_step_id",)),
    )
    .join_from(ETLStepRunEntity, ETLStepEntity)
    .order_by(ETLStepRunEntity.run_id.desc(), ETLStepRunEntity.ordering)  # type: ignore
)

current_run_view_statement = run_view_statement.where(
    ETLStepRunEntity.run_id == select(func.max(RunEntity.id)).scalar_subquery()  # type: ignore
)
failed_run_view_statement = run_view_statement.where(
    ETLStepRunEntity.status.in_(("failed", "running"))  # type: ignore
)


class CurrentRun(BaseEntity, registry=meta_registry):
    __table__ = create_view(
        "current_run",
        current_run_view_statement.subquery(),
        META_SCHEMA,
        meta_registry.metadata,
    )


class FailedRuns(BaseEntity, registry=meta_registry):
    __table__ = create_view(
        "failed_run",
        failed_run_view_statement.subquery(),
        META_SCHEMA,
        meta_registry.metadata,
    )


etl_steps_to_run_stmt = text(
    f"""
select
    distinct
    etl_step.name, cr.etl_step_id, coalesce(gr_bad.status::text, 'never run') as last_status, gr_bad.error
from
    {META_SCHEMA}.etl_step_run cr
join {META_SCHEMA}.etl_step etl_step on
    etl_step.id = cr.etl_step_id
left join {META_SCHEMA}.etl_step_run gr_completed on
    cr.etl_step_id = gr_completed.etl_step_id
    and gr_completed.status = 'completed'
left join {META_SCHEMA}.etl_step_run gr_bad on
    cr.etl_step_id = gr_bad.etl_step_id
    and gr_bad.status = 'failed'
where
    cr.status != 'completed'
    and cr.run_id = (
    select
        max(id)
    from
        {META_SCHEMA}.run)
    and (gr_completed.run_id is null
        or
     gr_bad.run_id>gr_completed.run_id)
"""
)


class ETLStepsToRun(BaseEntity, registry=meta_registry):
    __table__ = create_view(
        "etl_steps_to_run",
        etl_steps_to_run_stmt,
        META_SCHEMA,
        meta_registry.metadata,
        [
            Column("name", AutoString()),
            Column("etl_step_id", GUID()),
            Column("last_status", AutoString()),
            Column("error", AutoString()),
        ],
    )



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Objects related to the running of Models and ETLSteps."""
import asyncio
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from os import cpu_count
from time import time
from traceback import format_exc
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union
from uuid import UUID

import psutil
from psycopg import AsyncConnection
from psycopg.rows import dict_row
from psycopg_pool import AsyncConnectionPool
from pydasher import hasher
from sqlalchemy.future import Engine
from sqlmodel import Session

from dbgen.core.base import encoders
from dbgen.core.dashboard import BarNames, Dashboard
from dbgen.core.etl_step import ETLStep
from dbgen.core.metadata import ETLStepRunEntity, Repeats, RunEntity, Status
from dbgen.core.node.extract import Extract
from dbgen.core.node.query import BaseQuery
from dbgen.core.run.utilities import BaseETLStepExecutor
from dbgen.exceptions import DBgenExternalError, TransformerError
from dbgen.utils.log import setup_logger
from dbgen.utils.typing import NAMESPACE_TYPE, ROWS_TO_LOAD_TYPE

if TYPE_CHECKING:
    from asyncio.events import AbstractEventLoop

# Useful types
TRANSFORM_RETURN_TYPE = Union[
    Tuple[None, None, None, int, str], Tuple[list, Dict[str, Dict[UUID, dict]], int, int, None]
]

# TODO Add data to the Async Run object to minimize data passing around1
# TODO Refactor the methods to reduce verbosity and increase clarity
# TODO Type annottate queues
# TODO make return types clearer (dataclas?) to reduce verbosity
# TODO update gen_run during the run
class AsyncETLStepExecutor(BaseETLStepExecutor):
    def execute(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        meta_session: Session,
        run_id: Optional[int],
        dashboard: Optional[Dashboard],
        etl_step_run: ETLStepRunEntity,
    ):
        start = time()
        batch_size = self.run_config.batch_size or self.etl_step.batch_size or 1000
        (
            inputs_extracted,
            unique_inputs,
            inputs_processed,
            inputs_skipped,
            rows_inserted,
            rows_updated,
            memory_usage,
            exc,
        ) = asyncio.run(
            self.main(
                self.etl_step,
                main_dsn=str(main_engine.url),
                meta_dsn=str(meta_engine.url),
                batch_size=batch_size,
                dashboard=dashboard,
            )
        )
        if exc:
            etl_step_run.status = Status.failed
            etl_step_run.error = str(exc)
            self._logger.error(
                f"Encountered Error running etl_step {self.etl_step.name}({self.etl_step.uuid})."
            )
            run = meta_session.get(RunEntity, run_id)
            assert run
            run.errors = run.errors + 1 if run.errors else 1
            meta_session.commit()
            meta_session.close()
            return 1

        etl_step_run.status = Status.completed
        etl_step_run.inputs_extracted = inputs_extracted
        etl_step_run.unique_inputs = unique_inputs
        etl_step_run.inputs_skipped = inputs_skipped
        etl_step_run.inputs_processed = inputs_processed
        etl_step_run.rows_updated = rows_updated
        etl_step_run.rows_inserted = rows_inserted
        etl_step_run.memory_usage = memory_usage
        etl_step_run.runtime = round(time() - start, 3)
        self._logger.info(
            f"Finished running etl_step {self.etl_step.name}({self.etl_step.uuid}) in {etl_step_run.runtime}(s)."
        )
        self._logger.info(f"Inserted approximately {etl_step_run.rows_inserted} rows")
        self._logger.info(f"Updated approximately {etl_step_run.rows_updated} rows")
        meta_session.commit()
        meta_session.close()
        return

    async def main(
        self,
        etl_step: ETLStep,
        main_dsn: str,
        meta_dsn: str,
        batch_size: int,
        dashboard: Optional[Dashboard],
    ):
        # Initialize multiprocessing start method
        conn_pool = AsyncConnectionPool(main_dsn, name='test', min_size=4)
        meta_conn_pool = AsyncConnectionPool(meta_dsn, name='test', min_size=4)

        await conn_pool.check()
        loop = asyncio.get_running_loop()
        # Remove the stored functions on the etl_step for pickling
        etl_step.remove_stored_func()
        # Initialize the queues and type them
        transform_queue: asyncio.Queue[Tuple[Optional[UUID], Optional[NAMESPACE_TYPE]]] = asyncio.Queue()
        tform_results: asyncio.Queue[asyncio.Future[TRANSFORM_RETURN_TYPE]] = asyncio.Queue()
        load_queue: asyncio.Queue[Tuple[List[UUID], ROWS_TO_LOAD_TYPE, int]] = asyncio.Queue()
        repeats_queue: asyncio.Queue[Set[UUID]] = asyncio.Queue()

        rows_inserted, rows_loaded = 0, 0
        max_memory = None
        if dashboard:
            dashboard.add_etl_progress_bars(run_async=True)
        try:
            context = multiprocessing.get_context("spawn")
            with ProcessPoolExecutor(
                cpu_count(),
                mp_context=context,
                initializer=setup_logger,  # type: ignore
                initargs=(self.run_config.log_level, self.run_config.log_level),
            ) as executor:
                mem_usage = asyncio.create_task(self.memory_usage(executor))
                routines = (
                    self.extractor(
                        etl_step.extract,
                        transform_queue,
                        load_queue,
                        main_dsn,
                        batch_size=batch_size,
                        dashboard=dashboard,
                        etl_step_id=etl_step.uuid,
                        retry=self.run_config.retry,
                    ),
                    self.transformer(
                        etl_step,
                        transform_queue,
                        tform_results,
                        loop,
                        batch_size=batch_size,
                        dashboard=dashboard,
                        executor=executor,
                    ),
                    self.transformer_results(
                        tform_results,
                        load_queue,
                        dashboard=dashboard,
                    ),
                    self.loader(etl_step, load_queue, repeats_queue, conn_pool, dashboard=dashboard),
                    self.repeat_loader(repeats_queue, meta_conn_pool, etl_step.uuid),
                    self.set_length(etl_step.extract, dashboard, conn_pool),
                )
                tasks = map(asyncio.create_task, routines)
                results = await asyncio.gather(*tasks)
                self._logger.debug('Gathered tasks returned')
                self._logger.debug('Shutting down the executor...')
                mem_usage.cancel()
                max_memory = await mem_usage
            self._logger.debug('Executor shutdown')
        except (Exception, DBgenExternalError) as exc:
            self._logger.error('Uncaught exception found!')
            self._logger.exception(exc, exc_info=exc)
            self._logger.debug('Shutting down tasks')
            for task in tasks:
                task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            self._logger.debug('All tasks successfully shutdown')
            return None, None, None, None, None, None, None, format_exc(chain=False)
        finally:
            await conn_pool.close()

        inputs_extracted, unique_inputs, inputs_processed = results[0]
        inputs_skipped = results[2]
        rows_inserted, rows_loaded = results[3]
        return (
            inputs_extracted,
            unique_inputs,
            inputs_processed,
            inputs_skipped,
            rows_inserted,
            rows_loaded,
            max_memory,
            None,
        )

    async def extractor(
        self,
        extract: Extract,
        queue: 'asyncio.Queue[Tuple[Optional[UUID], Optional[NAMESPACE_TYPE]]]',
        load_queue: 'asyncio.Queue',
        async_dsn: str,
        batch_size,
        dashboard: Optional[Dashboard],
        etl_step_id: UUID,
        retry: bool,
    ) -> Tuple[int, int, int]:
        """Take a query and param and stream the outputs to the queue."""
        logger = self._logger.getChild('extractor')
        unique_inputs, inputs_extracted, inputs_processed = 0, 0, 0
        with extract:
            if not isinstance(extract, BaseQuery):
                for i, row in enumerate(extract.extract()):
                    if dashboard:
                        dashboard.advance_bar(BarNames.EXTRACTED, advance=1)
                    is_repeat, input_hash = self._check_repeat(row, etl_step_id)
                    # increment unique inputs and extracted inputs
                    inputs_extracted += 1
                    unique_inputs += 1 if not is_repeat else 0
                    if not is_repeat or retry:
                        await queue.put((input_hash, {extract.hash: extract.process_row(row)}))
                        inputs_processed += 1
                    else:
                        continue
                    if i % batch_size == 0:
                        await asyncio.sleep(0.05)
                await queue.put((None, None))
                # Start the bars with the fully extracted total
                if dashboard:
                    dashboard.set_total(i)
            else:
                async with await AsyncConnection.connect(async_dsn) as conn:
                    async with conn.cursor(row_factory=dict_row) as cursor:
                        result = await cursor.execute(extract.compiled_query, extract.params)
                        i = 0
                        async for row in result:
                            logged = False
                            while (queue.qsize() / batch_size) < (load_queue.qsize() - 100):
                                if not logged:
                                    self._logger.info('Waiting for loader to catch up')
                                    logged = True
                                await asyncio.sleep(1)

                            is_repeat, input_hash = self._check_repeat(row, etl_step_id)
                            # increment unique inputs and extracted inputs
                            inputs_extracted += 1
                            unique_inputs += 1 if not is_repeat else 0
                            if not is_repeat or retry:
                                await queue.put((input_hash, {extract.hash: extract.process_row(row)}))
                            else:
                                continue
                            if dashboard:
                                dashboard.advance_bar(BarNames.EXTRACTED, advance=1)
                            if i % batch_size == 0:
                                await asyncio.sleep(0.05)
                            i += 1

                        await queue.put((None, None))
        logger.debug('Extraction Finished')
        return inputs_extracted, unique_inputs, inputs_processed

    async def transformer(
        self,
        etl_step: ETLStep,
        transform_queue: asyncio.Queue,
        transformed_queue: asyncio.Queue,
        loop: 'AbstractEventLoop',
        batch_size: int,
        dashboard: Optional[Dashboard],
        executor,
    ):
        batch = []
        logger = self._logger.getChild('transformer')
        pending_tasks: Set[asyncio.Future[TRANSFORM_RETURN_TYPE]] = set()
        while True:
            # get row from producer queue
            input_hash, record = await transform_queue.get()
            # add to the batch
            if record is not None:
                batch.append((input_hash, record))
            # if batch is right size launch the transform
            if len(batch) >= batch_size or record is None:
                # Run transform on the cpu_pool for parallelization
                if len(pending_tasks) >= executor._max_workers:
                    logger.debug('waiting for max workers')
                    _, pending_tasks = await asyncio.wait(pending_tasks, return_when=asyncio.FIRST_COMPLETED)
                task = loop.run_in_executor(executor, etl_step.transform_batch, batch, self.run_config)
                pending_tasks.add(task)  # type: ignore
                # add to currently running tasks
                await transformed_queue.put(task)
                logger.debug(f"Number of items in Transformed Queue: {transformed_queue.qsize()}")
                if dashboard:
                    dashboard.advance_bar(BarNames.LAUNCHED, advance=len(batch))
                batch = []
            # if None is received from extract queue stop the loop
            if record is None:
                await transformed_queue.put(None)
                break
        logger.debug('Launching Finished')

    async def transformer_results(
        self,
        transformed_queue: 'asyncio.Queue[asyncio.Future[TRANSFORM_RETURN_TYPE]]',
        load_queue: asyncio.Queue,
        dashboard: Optional[Dashboard],
        timeout: float = 0.05,
    ):
        task_set: Set[asyncio.Future[TRANSFORM_RETURN_TYPE]] = set()
        logger = self._logger.getChild('results')
        inputs_skipped = 0
        while True:
            logger.debug('waiting for row')
            task = await transformed_queue.get()
            if task is None:
                break
            logger.debug('got row')
            task_set.add(task)
            logger.debug('waiting for tasks to finish')

            done_set, task_set = await asyncio.wait(
                task_set, return_when=asyncio.FIRST_COMPLETED, timeout=timeout
            )
            logger.debug(f'{len(done_set)} tasks finished')
            logger.debug(f'{len(task_set)} tasks pending')
            for task in done_set:
                processed_hashes, rows_to_load, update, skipped, tb = await task
                inputs_skipped += skipped
                if tb is not None:
                    raise TransformerError(tb)
                if dashboard and update:
                    dashboard.advance_bar(BarNames.TRANSFORMED, advance=update)
                await load_queue.put((processed_hashes, rows_to_load, update))
                logger.debug(f"Number of items in Load Queue: {load_queue.qsize()}")
        if task_set:
            logger.debug(f'queue finished waiting for remaining {len(task_set)} task(s) to finish')
            try:
                for task in asyncio.as_completed(task_set):
                    processed_hashes, rows_to_load, update, skipped, tb = await task
                    inputs_skipped += skipped
                    if tb is not None:
                        raise TransformerError(tb)
                    if dashboard:
                        dashboard.advance_bar(BarNames.TRANSFORMED, advance=update)
                    await load_queue.put((processed_hashes, rows_to_load, update))
                    logger.debug('pushed Task')
            except asyncio.TimeoutError:
                logger.error('While waiting for final items in queue to be transformed timeout was reached!')
                done_set, task_set = await asyncio.wait(
                    task_set, return_when=asyncio.FIRST_EXCEPTION, timeout=5
                )
                logger.error(f'{len(done_set)} are done!')
                logger.error(f'Cancelling remaining {len(task_set)}')
                for task in task_set:
                    task.cancel()
                await asyncio.wait_for(asyncio.gather(*task_set, return_exceptions=True), timeout=5)
        await load_queue.put((None, None, None))
        logger.debug('Results Finished')
        return inputs_skipped

    @staticmethod
    async def merge_rows(current_rows, new_rows) -> dict:
        for load_hash, rows in new_rows.items():
            updated_rows = current_rows.get(load_hash, {})
            updated_rows.update(rows)
            current_rows[load_hash] = updated_rows
        return current_rows

    def _check_repeat(self, extracted_dict, etl_step_uuid: UUID) -> Tuple[bool, UUID]:
        # Convert Row to a dictionary so we can hash it for repeat-checking
        input_hash = UUID(hasher((etl_step_uuid, extracted_dict), encoders=encoders))
        # If the input_hash has been seen and we don't have retry=True skip row
        is_repeat = input_hash in self._old_repeats or input_hash in self._new_repeats
        return (is_repeat, input_hash)

    async def loader(
        self,
        etl_step: ETLStep,
        load_queue: 'asyncio.Queue[Tuple[List[UUID], dict, int]]',
        repeat_queue: 'asyncio.Queue[Set[UUID]]',
        conn_pool: AsyncConnectionPool,
        dashboard: Optional[Dashboard],
    ):
        rows_inserted = 0
        rows_updated = 0
        logger = self._logger.getChild('loader')
        while True:
            logger.debug('loader waiting for row')
            rows_to_load: ROWS_TO_LOAD_TYPE = {}
            number_of_rows = 0
            number_of_batches = 0
            processed_hashes: Set[UUID] = set()
            while True:
                new_processed_hashes, record, update = await load_queue.get()
                if record is None:
                    break
                number_of_batches += 1
                number_of_rows += update
                processed_hashes = processed_hashes.union(new_processed_hashes)
                await self.merge_rows(rows_to_load, record)
                if load_queue.empty() or number_of_batches > 60:
                    logger.debug('queue is empty moving on')
                    break
            if rows_to_load:
                async with conn_pool.connection() as connection:
                    for load in etl_step.loads:
                        rows = rows_to_load[load.hash]
                        logger.debug(f'Loading into {load}')
                        rows_modified = await load._async_load(rows, connection, etl_step.uuid)
                        if load.insert:
                            rows_inserted += rows_modified
                        else:
                            rows_updated += rows_modified
            if dashboard:
                dashboard.advance_bar(BarNames.LOADED, advance=number_of_rows)
            await repeat_queue.put(processed_hashes)
            if record is None:
                await repeat_queue.put(None)
                break
        logger.debug('Loading Finished')
        return rows_inserted, rows_updated

    async def repeat_loader(
        self, repeat_queue: 'asyncio.Queue[Set[UUID]]', conn_pool: AsyncConnectionPool, etl_step_id: UUID
    ):
        """Load hash of repeated rows into meta-database"""
        logger = self._logger.getChild('repeat_checker')
        while True:
            logger.debug('repeat checker waiting for row')
            repeats: Set[UUID] = set()
            while True:
                new_repeats = await repeat_queue.get()
                if new_repeats is not None:
                    repeats = repeats.union(new_repeats)
                if repeat_queue.empty() or new_repeats is None:
                    break
            logger.debug('getting ready to load repeats')
            if repeats:
                logger.debug(f'Loading {len(repeats)}repeats')
                await self._load_repeats(conn_pool, repeats, etl_step_id)
            if new_repeats is None:
                break
        logger.debug('Repeat Checking Finished')

    async def _load_repeats(
        self,
        conn_pool: AsyncConnectionPool,
        repeats: Set[UUID],
        etl_step_id: UUID,
    ) -> None:
        rows = {input_hash: (etl_step_id,) for input_hash in repeats - self._old_repeats}
        async with conn_pool.connection() as connection:
            await Repeats._async_quick_load(connection, rows, column_names=["etl_step_id"])
        self._old_repeats = self._old_repeats.union(repeats)

    async def set_length(
        self, extract: Extract, dashboard: Optional[Dashboard], conn_pool: AsyncConnectionPool
    ):
        logger = self._logger.getChild('set_length')
        if isinstance(extract, BaseQuery):
            logger.debug('querying for the query length')
            async with conn_pool.connection() as aconn:
                logger.debug('got connection')
                total = await extract._async_length(connection=aconn)

            logger.debug('got query length')
        else:
            total = extract.length()
        if total is not None and dashboard:
            dashboard.set_total(total)
        return total

    async def memory_usage(self, executor: ProcessPoolExecutor, refresh_per_second: int = 1):
        """Monitor for the memory usage of the async run and the child processes"""
        logger = self._logger.getChild('memory')
        get_memory = lambda pid=None: psutil.Process(pid).memory_info().rss / (1024 * 1024)
        max_memory = 0
        try:
            while True:
                async_memory_usage = get_memory()
                total_memory_usage = async_memory_usage
                pids = []
                if executor._processes:
                    pids = list(executor._processes.keys())
                    total_memory_usage += sum(map(get_memory, pids))
                max_memory = max(total_memory_usage, max_memory)
                logger.debug(
                    f'Memory Usage ({len(pids)} child processes): Main = {async_memory_usage:3.1f} MB, Total = {total_memory_usage:3.1f} MB'
                )
                await asyncio.sleep(1 / refresh_per_second)

        except asyncio.CancelledError:
            logger.info(f'Max memory used {max_memory:3.1f} MB')
            logger.debug('Memory Usage Finished')
        return max_memory



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Welcome to DBgen!"""

__author__ = "Michael Statt"
__email__ = "michael.statt@modelyst.io"
__maintainer__ = "Michael Statt"
__maintainer_email__ = "michael.statt@modelyst.io"
__version__ = "1.0.0a4"


from sqlalchemy.orm import registry

# External packages that we reexport for convenience
from sqlmodel import select

from dbgen.core.args import Constant
from dbgen.core.decorators import transform
from dbgen.core.entity import BaseEntity, Entity
from dbgen.core.etl_step import ETLStep
from dbgen.core.func import Environment, Import
from dbgen.core.model import Model
from dbgen.core.model_settings import BaseModelSettings
from dbgen.core.node.extract import Extract
from dbgen.core.node.query import Query
from dbgen.core.node.transforms import PythonTransform
from dbgen.utils.typing import IDType



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any
from typing import Generator as GenType
from typing import List, Mapping, Optional

from sqlalchemy.future import Engine
from sqlmodel import Session, func, select

from dbgen.core.metadata import ETLStepEntity, ETLStepRunEntity, RunEntity


def get_runs(
    run_id: Optional[int],
    meta_engine: "Engine",
    all_runs: bool = True,
    accepted_statuses: List[str] = None,
    last: Optional[int] = 1,
) -> GenType[Mapping[str, Any], None, None]:
    statement = select(ETLStepEntity.name, *ETLStepRunEntity.__table__.c).join_from(ETLStepRunEntity, ETLStepEntity).order_by(ETLStepRunEntity.run_id.desc(), ETLStepRunEntity.ordering)  # type: ignore
    if accepted_statuses:
        statement = statement.where(ETLStepRunEntity.status.in_(accepted_statuses))  # type: ignore
    if not all_runs:
        statement = statement.where(
            ETLStepRunEntity.run_id > select(func.max(RunEntity.id)).scalar_subquery() - last  # type: ignore
        )
    with Session(meta_engine) as session:
        result = session.exec(statement).mappings()
        yield from result



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from datetime import date, datetime
from pathlib import Path, PosixPath
from typing import Any, Callable, ClassVar, Dict, Tuple, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo
from pydantic.main import ModelMetaclass
from pydasher import HashMixIn

_T = TypeVar("_T")


def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[type, Callable[..., Any]], ...] = (()),
) -> Callable[[_T], _T]:
    return lambda a: a


@__dataclass_transform__(kw_only_default=True, field_descriptors=(Field, FieldInfo))
class BaseMeta(ModelMetaclass):
    pass


encoders = {
    UUID: lambda x: str(x),
    datetime: lambda x: x.isoformat(),
    date: lambda x: x.isoformat(),
    Path: lambda x: str(x),
    PosixPath: lambda x: str(x),
}


class Base(HashMixIn, BaseModel, metaclass=BaseMeta):
    """Common methods shared by many DbGen objects."""

    _logger: ClassVar[logging.Logger]
    _logger_name: ClassVar[
        Union[Callable[["Base", Dict[str, Any]], str], str]
    ] = lambda cls, _: cls.canonical_name()

    def __new__(cls, *_args, **kwargs):
        # Call  cls._logger_name function with class and kwargs as arguments
        # to allow dynamic logger names (which is great for table by table loggers)
        logger_name = cls._logger_name(cls, kwargs) if callable(cls._logger_name) else cls._logger_name
        cls._logger = logging.getLogger(logger_name)
        return super().__new__(cls)

    class Config:
        """Pydantic Config"""

        json_encoders = encoders
        copy_on_model_validation = False

    @classmethod
    def canonical_name(cls) -> str:
        return cls.__module__ + "." + cls.__qualname__
